"""Tests for mdpo text utilities."""

import math

import pytest

from mdpo.text import (
    min_not_max_chars_in_a_row,
    parse_escaped_pair,
    parse_strint_0_inf,
    parse_wrapwidth_argument,
    removeprefix,
    removesuffix,
)


@pytest.mark.parametrize(
    ('char', 'text', 'expected_result'), (
        ('`', 'hello `` ``` a `` `', 4),
        ('`', 'hello `` ``` a ```` `', 5),
        ('`', 'hello `` ``` a `` `````', 1),
        ('`', 'hello ` `` ``` a `` `````', 4),

        # as default, 1
        ('`', 'hello', 1),

        # does not work for multiple characters (returns default)
        ('``', 'hello `` ``` a `` `````', 1),
    ),
)
def test_min_not_max_chars_in_a_row(char, text, expected_result):
    assert min_not_max_chars_in_a_row(char, text) == expected_result


@pytest.mark.parametrize(
    ('text', 'expected_key', 'expected_value', 'expected_error'),
    (
        ('foo:bar', 'foo', 'bar', None),
        ('foo:Bar:', 'foo', 'Bar:', None),
        ('foo:   bar:', 'foo', 'bar:', None),
        ('foo: \n  bar:', 'foo', 'bar:', None),
        ('foo:Bar\\:', 'foo', 'Bar\\:', None),
        ('\\:foo:bar', ':foo', 'bar', None),
        (r'foo\\:Bar:baz', 'foo:Bar', 'baz', None),
        (r'foo\\:Bar:baz\\', 'foo:Bar', r'baz\\', None),
        # : at the beginning means escaped
        (r':foo\\:Bar:baz\\', ':foo:Bar', r'baz\\', None),
        ('foo', None, None, ValueError),
        (':', None, None, ValueError),
        ('', None, None, ValueError),
    ),
)
def test_parse_escaped_pair(
    text,
    expected_key,
    expected_value,
    expected_error,
):
    if expected_error:
        with pytest.raises(expected_error):
            parse_escaped_pair(text)
    else:
        key, value = parse_escaped_pair(text)
        assert key == expected_key
        assert value == expected_value


@pytest.mark.parametrize(
    ('value', 'expected_value', 'expected_error'),
    (
        ('1', 1, None),
        ('1.1', 1, None),
        (-1, math.inf, None),
        (-1.1, math.inf, None),
        ('-1', math.inf, None),
        ('-1.1', math.inf, None),
        (0, math.inf, None),
        (-0, math.inf, None),
        ('a', None, ValueError),
        ('inf', math.inf, None),
        ('-inf', math.inf, None),
        (-1.1, float('inf'), None),
        ('-1', float('inf'), None),
    ),
)
def test_parse_strint_0_inf(value, expected_value, expected_error):
    if expected_error:
        with pytest.raises(expected_error):
            parse_strint_0_inf(value)
    else:
        assert parse_strint_0_inf(value) == expected_value


@pytest.mark.parametrize(
    ('value', 'prefix', 'expected_value'),
    (
        ('foo', 'fo', 'o'),
        ('bar', 'fo', 'bar'),
    ),
)
def test_removeprefix(value, prefix, expected_value):
    assert removeprefix(value, prefix) == expected_value


@pytest.mark.parametrize(
    ('value', 'suffix', 'expected_value'),
    (
        ('foo', 'oo', 'f'),
        ('bar', 'fo', 'bar'),
    ),
)
def test_removesuffix(value, suffix, expected_value):
    assert removesuffix(value, suffix) == expected_value


@pytest.mark.parametrize('value', (0, 80, 'inf', 'invalid'))
def test_parse_wrapwidth_argument(value):
    if value == 'invalid':
        with pytest.raises(ValueError) as exc:
            parse_wrapwidth_argument(value)
        assert str(exc.value) == (
            f'Invalid value \'{value}\' for wrapwidth argument.'
        )
        return

    assert parse_wrapwidth_argument(value) == (
        float('inf') if value == 0 else float(value)
    )



"""PO files related stuff."""

import glob

import polib

from mdpo.io import filter_paths
from mdpo.polib import poentry__cmp__


def po_escaped_string(chars):
    r"""Convenience function that prepends a ``\`` character to a string.

    This is used to escape values inside strings wrapped for the returned
    character.

    Args:
        chars (str): Value for which first character will be escaped.

    Returns:
        str: First character of passed string with ``\`` character prepended.
    """
    return f'\\{chars[0]}'


def find_entry_in_entries(entry, entries, **kwargs):
    """Returns an equal entry in a set of :py:class:`polib.POEntry` entries.

    Finds the first :py:class:`polib.POEntry` instance in the iterable
    ``entries`` that is equal, according to its ``__cmp__`` method, to
    the :py:class:`polib.POEntry` instance passed as ``entry`` argument.

    Args:
        entry (:py:class:`polib.POEntry`): Entry to search for.
        entries (list): Entries to search against.
        **kwargs: Keyword arguments passed to :py:meth:`polib.POEntry.__cmp__`.

    Returns:
        :py:class:`polib.POEntry`: Entry passed in ``entry`` argument if an
        equal entry has been found in ``entries`` iterable, otherwise ``None``.
    """
    response = None
    for compared_entry in entries:
        if poentry__cmp__(entry, compared_entry, **kwargs) == 0:
            response = compared_entry
            break
    return response


def mark_not_found_entries_as_obsoletes(
    pofile,
    entries,
):
    """Marks entries in a PO file obsoletes if are not in a set of entries.

    If an entry of the PO file is found in the set of entries, will be marked
    as no obsolete.

    Args:
        pofile (:py:class:`polib.POFile`): PO file in which the missing entries
            will be marked as obsoletes.
        entries (list): Entries to search against.
    """
    for entry in pofile:
        if not find_entry_in_entries(
            entry,
            entries,
            compare_occurrences=False,
        ):
            _equal_not_obsolete_found = find_entry_in_entries(
                entry,
                entries,
                compare_obsolete=False,
                compare_occurrences=False,
            )
            if _equal_not_obsolete_found:
                pofile.remove(entry)
            else:
                entry.obsolete = True
        else:
            entry.obsolete = False


def remove_not_found_entries(pofile, entries):
    """Removes entries in a PO file if are not in a set of entries.

    Args:
        pofile (:py:class:`polib.POFile`): PO file for which the missing
            entries will be removed.
        entries (list): Entries to search against.
    """
    for entry in pofile:
        if not find_entry_in_entries(
            entry,
            entries,
            compare_occurrences=False,
        ):
            _equal_not_obsolete_found = find_entry_in_entries(
                entry,
                entries,
                compare_obsolete=False,
                compare_occurrences=False,
            )
            if not _equal_not_obsolete_found:
                pofile.remove(entry)


def pofiles_to_unique_translations_dicts(pofiles):
    """Extracts unique translations from a set of PO files.

    Given multiple pofiles, extracts translations (those messages with non
    empty msgstrs) into two dictionaries, a dictionary for translations
    with contexts and other without them.

    Args:
        pofiles (list): List of :py:class:`polib.POFile` objects.

    Returns:
        tuple: dictionaries with translations.
    """
    translations, translations_with_msgctxt = ({}, {})
    for pofile in pofiles:
        for entry in pofile:
            if entry.msgctxt:
                if entry.msgctxt not in translations_with_msgctxt:
                    translations_with_msgctxt[entry.msgctxt] = {}
                translations_with_msgctxt[
                    entry.msgctxt
                ][entry.msgid] = entry.msgstr
            else:
                translations[entry.msgid] = entry.msgstr
    return (translations, translations_with_msgctxt)


def paths_or_globs_to_unique_pofiles(pofiles_globs, ignore, po_encoding=None):
    """Converts any path, paths or glob to :py:class:`polib.POFile` objects.

    Args:
        pofiles_globs (str, list): Can be a path, a glob, multiples paths
            or multiples globs.
        ignore (list): Paths to ignore.
        po_encoding (str): Encoding used reading the PO files.

    Returns:
        set: Unique set of :py:class:`polib.POFile` objects.
    """
    if isinstance(pofiles_globs, str):
        pofiles_globs = [pofiles_globs]

    _po_filepaths, pofiles = ([], [])

    for pofiles_glob in pofiles_globs:
        for po_filepath in filter_paths(
            glob.glob(pofiles_glob),
            ignore_paths=ignore,
        ):
            if po_filepath not in _po_filepaths:
                pofiles.append(
                    polib.pofile(po_filepath, encoding=po_encoding),
                )
                _po_filepaths.append(po_filepath)

    return pofiles



#!/usr/bin/env python

"""md2po2md command line interface.

See :ref:`md2po2md CLI<cli:md2po2md>`.
"""

import argparse
import sys

from mdpo.cli import (
    add_command_alias_argument,
    add_common_cli_first_arguments,
    add_debug_option,
    add_encoding_arguments,
    add_extensions_argument,
    add_nolocation_option,
    add_pre_commit_option,
    parse_command_aliases_cli_arguments,
)
from mdpo.context import environ
from mdpo.md2po2md import markdown_to_pofile_to_markdown
from mdpo.md4c import DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS


DESCRIPTION = (
    'Translates Markdown files using PO files for a set of predefined language'
    ' codes creating multiple directories, one for each language.'
)


def build_parser():
    parser = argparse.ArgumentParser(description=DESCRIPTION, add_help=False)
    add_common_cli_first_arguments(parser, quiet=False)
    parser.add_argument(
        'input_paths_glob', metavar='GLOB', nargs='*',
        help='Glob to markdown input files to translate.'
             ' If not provided, will be read from STDIN.',
    )
    parser.add_argument(
        '-l', '--lang', dest='langs', default=[], action='append',
        help='Language codes used to create the output directories.'
             ' This argument can be passed multiple times.',
        metavar='LANG',
    )
    parser.add_argument(
        '-o', '--output', dest='output_paths_schema',
        required=True, type=str,
        help='Path schema for outputs, built using placeholders. There is a'
             ' mandatory placeholder for languages: \'{lang}\'; and one'
             ' optional for output basename: \'{basename}\'. For example,'
             ' for the schema \'locale/{lang}\', the languages \'es\' and'
             ' \'fr\' and a \'README.md\' as input, the next files will be'
             ' written: \'locale/es/README.po\', \'locale/es/README.md\','
             ' \'locale/fr/README.po\' and \'locale/fr/README.md\'.'
             ' Note that you can omit \'{basename}\', specifying a'
             ' directory for each language with \'locale/{lang}\' for this'
             ' example. Unexistent directories and files will be created, '
             ' so you don\'t have to prepare the output directories before'
             ' the execution.',
        metavar='PATH_SCHEMA',
    )
    add_nolocation_option(parser)
    add_extensions_argument(parser)
    add_command_alias_argument(parser)
    add_encoding_arguments(parser)
    add_debug_option(parser)
    add_pre_commit_option(parser)
    return parser


def parse_options(args=[]):
    parser = build_parser()
    if '-h' in args or '--help' in args:
        parser.print_help()
        sys.exit(1)
    opts, unknown = parser.parse_known_args(args)

    input_paths_glob = ''
    if not sys.stdin.isatty():
        input_paths_glob += sys.stdin.read().strip('\n')
    if isinstance(opts.input_paths_glob, list) and opts.input_paths_glob:
        input_paths_glob += opts.input_paths_glob[0]
    opts.input_paths_glob = input_paths_glob

    if opts.extensions is None:
        opts.extensions = DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS

    opts.command_aliases = parse_command_aliases_cli_arguments(
        opts.command_aliases,
    )

    return opts


def run(args=[]):
    exitcode = 0

    with environ(_MDPO_RUNNING='true'):
        opts = parse_options(args)

        kwargs = dict(
            extensions=opts.extensions,
            command_aliases=opts.command_aliases,
            debug=opts.debug,
            location=opts.location,
            po_encoding=opts.po_encoding,
            md_encoding=opts.md_encoding,
        )

        _saved_files_changed = markdown_to_pofile_to_markdown(
            opts.langs,
            opts.input_paths_glob,
            opts.output_paths_schema,
            _check_saved_files_changed=opts.check_saved_files_changed,
            **kwargs,
        )
        if opts.check_saved_files_changed and _saved_files_changed:
            exitcode = 1
    return exitcode


def main():
    sys.exit(run(args=sys.argv[1:]))  # pragma: no cover


if __name__ == '__main__':
    main()



"""mdpo HTML commands related utilities."""

from mdpo.text import removeprefix


COMMAND_SEARCH_REGEX = r'<\!\-\-\s{0,}([^\s]+)\s{0,}([\w\s]+)?\-\->'

MDPO_COMMANDS = [
    'context',
    'disable',
    'disable-next-line',
    'enable',
    'enable-next-line',
    'include',
    'include-codeblock',
    'include-codeblocks',
    'disable-codeblock',
    'disable-codeblocks',
    'translator',
]


def parse_mdpo_html_command(value):
    """Parses a mdpo HTML command inside a string value.

    This function is used by md2po implementation to discover the HTML commands
    used by mdpo to customize the extraction.

    Args:
        value (str): Text where will a command will be searched.

    Returns:
        tuple: Namename of the command (not including the ``"mdpo"`` prefix)
        and its value.
    """
    import re
    command_match = re.compile(COMMAND_SEARCH_REGEX).search(value)
    return command_match.groups() if command_match else (None, None)


def normalize_mdpo_command(value):
    """Normalizes a valid command and returns None if the command is invalid.

    This function works for lazy command validation (the user doesn't need
    to specify the ``mdpo-`` prefix used for command names).

    Args:
        value (str): Command to be normalized.

    Returns:
        str or None: Normalized command or None if is invalid.
    """
    if not value.startswith('mdpo-'):
        value = f'mdpo-{value}'
    return value if removeprefix(value, 'mdpo-') in MDPO_COMMANDS else None


def normalize_mdpo_command_aliases(command_aliases):
    """Validates and normalizes a mapping of mdpo command aliases.

    Args:
        command_aliases (dict): Aliases for mdpo commands, in the form
            ``{"<new-command-name>": "<mdpo-command>"}``.

    Raises:
        ValueError: if an mdpo command resolution doesn't exists.

    Returns:
        dict: Mapping validated and normalized.
    """
    response = {}
    for alias, mdpo_command in command_aliases.items():
        valid_mdpo_command = normalize_mdpo_command(mdpo_command)
        if not valid_mdpo_command:
            valid_mdpo_command_names_string = ', '.join(
                (f"'{cmd}'" for cmd in get_valid_mdpo_command_names()),
            )
            raise ValueError(
                f"Invalid mdpo command resolution '{mdpo_command}'."
                f' Valid values are: {valid_mdpo_command_names_string}',
            )
        response[alias] = valid_mdpo_command
    return response


def get_valid_mdpo_command_names():
    """Return valid mdpo commands names, with and without "mdpo-" prefix.

    Returns:
        list: Valid mdpo command names.
    """
    response = []
    for mdpo_command in MDPO_COMMANDS:
        response.append(f'mdpo-{mdpo_command}')
        response.append(mdpo_command)
    return response



import os
import subprocess
import tempfile


def pre_commit_run_all_files(cwd=os.getcwd()):
    return subprocess.run(
        ['pre-commit', 'run', '--all-files'],
        cwd=cwd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )


def test_md2po_pre_commit_hook(git_add_commit):
    with tempfile.TemporaryDirectory() as filesdir:
        pre_commit_config_path = os.path.join(
            filesdir,
            '.pre-commit-config.yaml',
        )
        readme_md_path = os.path.join(filesdir, 'README.md')
        readme_po_path = os.path.join(filesdir, 'README.po')

        with open(pre_commit_config_path, 'w') as f:
            f.write('''repos:
  - repo: https://github.com/mondeja/mdpo
    rev: master
    hooks:
      - id: md2po
        files: ^README\\.md
        args:
          - --po-filepath
          - README.po
''')

        with open(readme_md_path, 'w') as f:
            f.write('# Foo\n')
        with open(readme_po_path, 'w') as f:
            f.write('''#
msgid ""
msgstr ""

msgid "Foo"
msgstr ""
''')

        # first execution, is updated
        proc = subprocess.run(
            ['git', 'init'],
            cwd=filesdir,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert proc.returncode == 0

        git_add_commit('First commit', cwd=filesdir)

        proc = pre_commit_run_all_files(cwd=filesdir)
        assert proc.returncode == 0
        assert proc.stdout.decode('utf-8').splitlines()[-1].endswith('Passed')

        # second execution, is outdated
        with open(readme_md_path, 'a') as f:
            f.write('\nbar\n')

        git_add_commit('Second commit', cwd=filesdir)

        proc = pre_commit_run_all_files(cwd=filesdir)
        assert proc.returncode != 0

        assert proc.stdout.decode('utf-8').splitlines()[-1] == (
            '- files were modified by this hook'
        )

        with open(readme_po_path) as f:
            assert f.read() == '''#
msgid ""
msgstr ""

msgid "Foo"
msgstr ""

#: README.md:block 2 (paragraph)
msgid "bar"
msgstr ""
'''


def test_po2md_pre_commit_hook(git_add_commit):
    with tempfile.TemporaryDirectory() as filesdir:
        pre_commit_config_path = os.path.join(
            filesdir,
            '.pre-commit-config.yaml',
        )
        readme_src_md_path = os.path.join(filesdir, 'README.md')
        readme_dst_md_path = os.path.join(filesdir, 'README.es.md')
        readme_po_path = os.path.join(filesdir, 'README.po')

        with open(pre_commit_config_path, 'w') as f:
            f.write('''repos:
  - repo: https://github.com/mondeja/mdpo
    rev: master
    hooks:
      - id: po2md
        files: ^README\\.md
        args:
          - -p
          - README.po
          - -s
          - README.es.md
''')

        with open(readme_src_md_path, 'w') as f:
            f.write('# Foo\n')
        with open(readme_dst_md_path, 'w') as f:
            f.write('# Foo es\n')
        with open(readme_po_path, 'w') as f:
            f.write('''#
msgid ""
msgstr ""

msgid "Foo"
msgstr "Foo es"
''')

        # first execution, is updated
        proc = subprocess.run(
            ['git', 'init'],
            cwd=filesdir,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert proc.returncode == 0

        git_add_commit('First commit', cwd=filesdir)

        proc = pre_commit_run_all_files(cwd=filesdir)
        assert proc.returncode == 0
        assert proc.stdout.decode('utf-8').splitlines()[-1].endswith('Passed')

        # second execution, is outdated
        with open(readme_src_md_path, 'a') as f:
            f.write('\nbar\n')
        with open(readme_po_path, 'a') as f:
            f.write('\nmsgid "bar"\nmsgstr "bar es"\n')

        git_add_commit('Second commit', cwd=filesdir)

        proc = pre_commit_run_all_files(cwd=filesdir)
        assert proc.returncode != 0

        assert proc.stdout.decode('utf-8').splitlines()[-1] == (
            '- files were modified by this hook'
        )
        with open(readme_dst_md_path) as f:
            assert f.read() == '''# Foo es

bar es
'''


def test_mdpo2html_pre_commit_hook(git_add_commit):
    with tempfile.TemporaryDirectory() as filesdir:
        pre_commit_config_path = os.path.join(
            filesdir,
            '.pre-commit-config.yaml',
        )
        readme_html_path = os.path.join(filesdir, 'README.html')
        readme_html_es_path = os.path.join(filesdir, 'README.es.html')
        readme_po_path = os.path.join(filesdir, 'README.po')

        with open(pre_commit_config_path, 'w') as f:
            f.write('''repos:
  - repo: https://github.com/mondeja/mdpo
    rev: master
    hooks:
      - id: mdpo2html
        files: ^README\\.html
        args:
          - -p
          - README.po
          - -s
          - README.es.html
''')

        with open(readme_html_path, 'w') as f:
            f.write('<h1>Foo</h1>\n')
        with open(readme_html_es_path, 'w') as f:
            f.write('<h1>Foo es</h1>\n')
        with open(readme_po_path, 'w') as f:
            f.write('''#
msgid ""
msgstr ""

msgid "Foo"
msgstr "Foo es"
''')

        # first execution, is updated
        proc = subprocess.run(
            ['git', 'init'],
            cwd=filesdir,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert proc.returncode == 0

        git_add_commit('First commit', cwd=filesdir)

        proc = pre_commit_run_all_files(cwd=filesdir)
        assert proc.returncode == 0
        assert proc.stdout.decode('utf-8').splitlines()[-1].endswith('Passed')

        with open(readme_html_es_path) as f:
            assert f.read() == '<h1>Foo es</h1>\n'

        # second execution, is outdated
        with open(readme_html_path, 'a') as f:
            f.write('\n<p>bar</p>\n')
        with open(readme_po_path, 'a') as f:
            f.write('\nmsgid "bar"\nmsgstr "bar es"\n')

        git_add_commit('Second commit', cwd=filesdir)

        proc = pre_commit_run_all_files(cwd=filesdir)
        assert proc.returncode != 0

        assert proc.stdout.decode('utf-8').splitlines()[-1] == (
            '- files were modified by this hook'
        )
        with open(readme_html_es_path) as f:
            assert f.read() == '<h1>Foo es</h1>\n\n<p>bar es</p>\n'


def test_md2po2md_pre_commit_hook(git_add_commit):
    with tempfile.TemporaryDirectory() as filesdir:
        pre_commit_config_path = os.path.join(
            filesdir,
            '.pre-commit-config.yaml',
        )
        readme_md_path = os.path.join(filesdir, 'README.md')

        with open(pre_commit_config_path, 'w') as f:
            f.write(
                '''repos:
  - repo: https://github.com/mondeja/mdpo
    rev: master
    hooks:
      - id: md2po2md
        files: ^README\\.md
        args:
          - -l
          - es
          - -o
          - locale/{lang}
          - --no-location
''',
            )

        with open(readme_md_path, 'w') as f:
            f.write('# Foo\n')

        # first execution, files don't exist
        proc = subprocess.run(
            ['git', 'init'],
            cwd=filesdir,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert proc.returncode == 0

        git_add_commit('First commit', cwd=filesdir)

        proc = pre_commit_run_all_files(cwd=filesdir)
        assert proc.returncode == 1
        assert proc.stdout.decode('utf-8').splitlines()[-1] == '- exit code: 1'

        locale_dir = os.path.join(filesdir, 'locale')
        assert os.path.isdir(locale_dir)

        locale_es_dir = os.path.join(locale_dir, 'es')
        assert os.path.isdir(locale_es_dir)

        readme_md_es_path = os.path.join(locale_es_dir, 'README.md')
        readme_po_es_path = os.path.join(locale_es_dir, 'README.md.po')
        assert os.path.isfile(readme_md_es_path)
        assert os.path.isfile(readme_po_es_path)

        with open(readme_po_es_path) as f:
            assert f.read() == '''#
msgid ""
msgstr ""

msgid "Foo"
msgstr ""
'''

        with open(readme_md_es_path) as f:
            assert f.read() == '# Foo\n'

        # second execution, translation
        with open(readme_po_es_path, 'w') as f:
            f.write('''#
msgid ""
msgstr ""

msgid "Foo"
msgstr "Foo es"
''')

        git_add_commit('Second commit', cwd=filesdir)

        proc = pre_commit_run_all_files(cwd=filesdir)
        assert proc.returncode == 1
        assert proc.stdout.decode('utf-8').splitlines()[-1] == (
            '- files were modified by this hook'
        )

        with open(readme_md_es_path) as f:
            assert f.read() == '# Foo es\n'

        # third execution, is updated
        git_add_commit('Third commit', cwd=filesdir)

        proc = pre_commit_run_all_files(cwd=filesdir)
        assert proc.returncode == 0
        assert proc.stdout.decode('utf-8').splitlines()[-1].endswith('Passed')



"""Markdown files translator using PO files as reference."""

import md4c
import polib

from mdpo.command import (
    normalize_mdpo_command_aliases,
    parse_mdpo_html_command,
)
from mdpo.event import debug_events, raise_skip_event
from mdpo.io import save_file_checking_file_changed, to_file_content_if_is_file
from mdpo.md import (
    MarkdownSpanWrapper,
    escape_links_titles,
    parse_link_references,
)
from mdpo.md4c import DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS
from mdpo.po import (
    paths_or_globs_to_unique_pofiles,
    po_escaped_string,
    pofiles_to_unique_translations_dicts,
)
from mdpo.text import (
    min_not_max_chars_in_a_row,
    parse_wrapwidth_argument,
    removesuffix,
)


class Po2Md:
    __slots__ = {
        'pofiles',
        'output',
        'content',
        'extensions',
        'events',
        'disabled_entries',
        'translated_entries',
        'translations',
        'translations_with_msgctxt',
        'command_aliases',
        'wrapwidth',

        'bold_start_string',
        'bold_start_string_escaped',
        'bold_end_string',
        'bold_end_string_escaped',
        'italic_start_string',
        'italic_start_string_escaped',
        'italic_end_string',
        'italic_end_string_escaped',
        'code_start_string',
        'code_start_string_escaped',
        'code_end_string',
        'code_end_string_escaped',
        'link_start_string',
        'link_end_string',
        'wikilink_start_string',
        'wikilink_end_string',

        # internal config
        '_current_msgid',
        '_current_msgctxt',
        '_current_tcomment',
        '_current_line',
        '_outputlines',
        '_disable_next_line',
        '_disable',
        '_enable_next_line',
        '_enterspan_replacer',
        '_leavespan_replacer',
        '_saved_files_changed',

        # state
        '_inside_htmlblock',
        '_inside_codeblock',
        '_inside_indented_codeblock',
        '_inside_pblock',
        '_inside_liblock',
        '_inside_liblock_first_p',
        '_inside_hblock',
        '_inside_aspan',
        '_inside_codespan',
        '_inside_table',
        '_codespan_start_index',
        '_codespan_backticks',
        '_codespan_inside_current_msgid',
        '_inside_quoteblock',
        '_current_aspan_ref_target',
        '_current_aspan_href',
        '_current_aspan_text',
        '_current_imgspan',
        '_current_thead_aligns',
        '_aimg_title_inside_current_msgid',
        '_ul_marks',
        '_ol_marks',
        '_current_list_type',
        '_current_wikilink_target',
        '_link_references',
    }

    def __init__(self, pofiles, ignore=[], po_encoding=None, **kwargs):
        self.pofiles = paths_or_globs_to_unique_pofiles(
            pofiles,
            ignore,
            po_encoding=po_encoding,
        )

        self.extensions = kwargs.get(
            'extensions',
            DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS,
        )

        self.events = {}
        if 'events' in kwargs:
            for event_name, functions in kwargs['events'].items():
                self.events[event_name] = (
                    [functions] if callable(functions) else functions
                )
        if kwargs.get('debug'):
            for event_name, function in debug_events('po2md').items():
                if event_name not in self.events:
                    self.events[event_name] = []
                self.events[event_name].append(function)

        self._current_msgid = ''
        self._current_msgctxt = None
        self._current_tcomment = None
        self._current_line = ''
        self._outputlines = []

        self._disable_next_line = False
        self._disable = False
        self._enable_next_line = False
        self.disabled_entries = []
        self.translated_entries = []

        self.translations = None
        self.translations_with_msgctxt = None

        self.command_aliases = normalize_mdpo_command_aliases(
            kwargs.get('command_aliases', {}),
        )

        self.wrapwidth = (
            # infinte gives some undesired rendering
            (
                2 ** 24 if kwargs['wrapwidth'] in [float('inf'), 0]
                else parse_wrapwidth_argument(kwargs['wrapwidth'])
            ) if 'wrapwidth' in kwargs else 80
        )

        self._saved_files_changed = (
            False if kwargs.get('_check_saved_files_changed') else None
        )

        self.bold_start_string = kwargs.get('bold_start_string', '**')
        self.bold_start_string_escaped = po_escaped_string(
            self.bold_start_string,
        )

        self.bold_end_string = kwargs.get('bold_end_string', '**')
        self.bold_end_string_escaped = po_escaped_string(
            self.bold_end_string,
        )

        self.italic_start_string = kwargs.get('italic_start_string', '*')
        self.italic_start_string_escaped = po_escaped_string(
            self.italic_start_string,
        )

        self.italic_end_string = kwargs.get('italic_end_string', '*')
        self.italic_end_string_escaped = po_escaped_string(
            self.italic_end_string,
        )

        self.code_start_string = kwargs.get('code_start_string', '`')[0]
        self.code_start_string_escaped = po_escaped_string(
            self.code_start_string,
        )

        self.code_end_string = kwargs.get('code_end_string', '`')[0]
        self.code_end_string_escaped = po_escaped_string(
            self.code_end_string,
        )

        self._enterspan_replacer = {
            md4c.SpanType.STRONG.value: self.bold_start_string,
            md4c.SpanType.EM.value: self.italic_start_string,
            md4c.SpanType.CODE.value: self.code_start_string,
        }

        self._leavespan_replacer = {
            md4c.SpanType.STRONG.value: self.bold_end_string,
            md4c.SpanType.EM.value: self.italic_end_string,
            md4c.SpanType.CODE.value: self.code_end_string,
        }

        self.wikilink_start_string = '[['
        self.wikilink_end_string = ']]'

        if 'wikilinks' in self.extensions:
            self.wikilink_start_string = kwargs.get(
                'link_end_string',
                self.wikilink_start_string,
            )
            self.wikilink_end_string = kwargs.get(
                'link_end_string',
                self.wikilink_end_string,
            )

            self._enterspan_replacer[md4c.SpanType.WIKILINK.value] = \
                self.wikilink_start_string
            self._leavespan_replacer[md4c.SpanType.WIKILINK.value] = \
                self.wikilink_end_string

        self._inside_htmlblock = False
        self._inside_codeblock = False
        self._inside_indented_codeblock = False
        self._inside_hblock = False
        self._inside_pblock = False
        self._inside_liblock = False
        # first li block paragraph (li block "title")
        self._inside_liblock_first_p = False
        self._inside_table = False

        self._inside_codespan = False
        self._codespan_start_index = None
        self._codespan_backticks = None
        self._codespan_inside_current_msgid = False

        self._inside_quoteblock = False

        self._inside_aspan = False
        self._current_aspan_ref_target = None
        self._current_aspan_href = None
        self._current_aspan_text = ''

        self._link_references = None
        self._current_imgspan = {}

        # current table head alignments
        self._current_thead_aligns = []

        # if title are found in images of links for the current msgid
        # we need to escape them after translate it because ``polib.unescape``
        # removes the scapes
        self._aimg_title_inside_current_msgid = False

        # current UL marks by nesting levels
        self._ul_marks = []

        # [numerical iterm order, current delimitier] for OL blocks
        self._ol_marks = []

        # current lists type (list nesting): ['ul' or 'ol', [True, False]]
        # (second parameter is tasklist item or not)
        self._current_list_type = []

        self._current_wikilink_target = None

    def command(self, mdpo_command, comment, original_command):
        # raise 'command' event
        if raise_skip_event(
            self.events,
            'command',
            self,
            mdpo_command,
            comment,
            original_command,
        ):
            return

        if mdpo_command == 'mdpo-disable-next-line':
            self._disable_next_line = True
        elif mdpo_command == 'mdpo-disable':
            self._disable = True
        elif mdpo_command == 'mdpo-enable':
            self._disable = False
        elif mdpo_command == 'mdpo-enable-next-line':
            self._enable_next_line = True
        elif comment:
            if mdpo_command == 'mdpo-context':
                self._current_msgctxt = comment.rstrip()
            elif mdpo_command == 'mdpo-translator':
                self._current_tcomment = comment.rstrip()

    def _process_command(self, text):
        original_command, comment = parse_mdpo_html_command(text)
        if original_command is None:
            return

        try:
            command = self.command_aliases[original_command]
        except KeyError:  # not custom command
            command = original_command

        # process solved command
        self.command(command, comment, original_command)

    def _escape_translation(self, text):
        if self._aimg_title_inside_current_msgid:
            # escape '"' characters inside links and image titles
            text = escape_links_titles(text)
        return text

    def _translate_msgid(self, msgid, msgctxt, tcomment):
        try:
            if msgctxt:
                msgstr = self.translations_with_msgctxt[msgctxt][msgid]
            else:
                msgstr = self.translations[msgid]
        except KeyError:
            return msgid
        else:
            self.translated_entries.append(
                polib.POEntry(
                    msgid=msgid,
                    msgctxt=msgctxt,
                    msgstr=msgstr,
                    tcomment=tcomment,
                ),
            )
            return msgstr or msgid

    def _save_current_msgid(self):
        # raise 'msgid' event
        if raise_skip_event(
            self.events,
            'msgid',
            self,
            self._current_msgid,
            None,
            self._current_msgctxt,
            self._current_tcomment,
            [],
        ):
            return

        if (not self._disable and not self._disable_next_line) or \
                self._enable_next_line:
            translation = self._translate_msgid(
                self._current_msgid,
                self._current_msgctxt,
                self._current_tcomment,
            )
        else:
            translation = self._current_msgid
            self.disabled_entries.append(
                polib.POEntry(
                    msgid=translation,
                    msgstr='',
                    msgctxt=self._current_msgctxt,
                    tcomment=self._current_tcomment,
                ),
            )
        if self._inside_indented_codeblock:
            new_translation = ''
            for line in translation.splitlines():
                new_translation += f'    {line}\n'
            translation = new_translation
        else:
            first_line_width_diff = 0
            if self._inside_liblock or self._inside_quoteblock:
                first_line_width_diff = -2
                if self._inside_liblock:
                    if self._current_list_type[-1][0] == 'ul':
                        if self._current_list_type[-1][-1][-1] is True:
                            first_line_width_diff = -6
                    else:
                        first_line_width_diff = -3

            if not self._inside_codeblock:
                indent = ''
                if len(self._current_list_type) > 1:
                    indent = '   ' * len(self._current_list_type)

                translation = MarkdownSpanWrapper(
                    width=self.wrapwidth,
                    first_line_width=self.wrapwidth + first_line_width_diff,
                    indent=indent,
                    md4c_extensions=self.extensions,
                    code_start_string=self.code_start_string,
                    code_end_string=self.code_end_string,
                    italic_start_string_escaped=(
                        self.italic_start_string_escaped
                    ),
                    italic_end_string_escaped=self.italic_end_string_escaped,
                    code_start_string_escaped=self.code_start_string_escaped,
                    code_end_string_escaped=self.code_end_string_escaped,
                    wikilink_start_string=self.wikilink_start_string,
                    wikilink_end_string=self.wikilink_end_string,
                ).wrap(self._escape_translation(translation))

                if self._inside_hblock or self._inside_table:
                    translation = translation.rstrip('\n')

        if translation.rstrip('\n'):
            self._current_line += translation

        self._current_msgid = ''
        self._current_msgctxt = None
        self._current_tcomment = None

        self._disable_next_line = False
        self._enable_next_line = False

        self._codespan_inside_current_msgid = False
        self._aimg_title_inside_current_msgid = False

    def _save_current_line(self):
        # strip all spaces according to unicodedata database ignoring newlines,
        # see https://docs.python.org/3/library/stdtypes.html#str.splitlines
        self._outputlines.append(self._current_line.rstrip(' \v\x0b\f\x0c'))
        self._current_line = ''

    def enter_block(self, block, details):
        # raise 'enter_block' event
        if raise_skip_event(
            self.events,
            'enter_block',
            self, block,
            details,
        ):
            return

        if self._inside_quoteblock and (
                not self._current_line or self._current_line[0] != '>'
        ):
            self._current_line += '> '
        if block is md4c.BlockType.P:
            self._inside_pblock = True
        elif block is md4c.BlockType.CODE:
            self._inside_codeblock = True
            indent = ''

            if self._inside_liblock:
                self._save_current_msgid()
                if self._current_line:
                    self._save_current_line()
                indent += '   ' * len(self._current_list_type)

            if details['fence_char'] is not None:
                fence_chars = details['fence_char']*3
                self._current_line += f'{indent}{fence_chars}'
                if details['lang']:
                    self._current_line += details['lang'][0][1]
            else:
                self._inside_indented_codeblock = True

            if self._current_line:
                self._save_current_line()
        elif block is md4c.BlockType.H:
            self._inside_hblock = True
            hash_signs = '#' * details['level']
            self._current_line += f'{hash_signs} '
        elif block is md4c.BlockType.LI:
            if self._current_list_type[-1][0] == 'ol':
                # inside OL
                if len(self._ol_marks) > 1:
                    self._save_current_msgid()
                    if not self._ol_marks[-1][0]:
                        self._save_current_line()
                self._ol_marks[-1][0] += 1
                indent = '   ' * (len(self._current_list_type) - 1)
                self._current_line += f'{indent}1{self._ol_marks[-1][1]} '
                self._current_list_type[-1][-1].append(False)
            else:
                # inside UL
                indent = '   ' * (len(self._current_list_type) - 1)
                self._current_line += f'{indent}{self._ul_marks[-1]} '
                if details['is_task']:
                    mark = details['task_mark']
                    self._current_line += f'[{mark}] '
                self._current_list_type[-1][-1].append(details['is_task'])
            self._inside_liblock = True
            self._inside_liblock_first_p = True
        elif block is md4c.BlockType.UL:
            if self._current_list_type:
                self._save_current_msgid()
                self._save_current_line()
            self._current_list_type.append(['ul', []])
            self._ul_marks.append(details['mark'])
        elif block is md4c.BlockType.OL:
            self._current_list_type.append(['ol', []])
            self._ol_marks.append([0, details['mark_delimiter']])
        elif block is md4c.BlockType.HR:
            if not self._inside_liblock:
                self._current_line += '---\n\n'
            else:
                # inside lists, the separator '---' can't be used
                self._current_line += '***'
        elif block is md4c.BlockType.TR:
            self._current_line += '   ' * len(self._current_list_type)
            if self._current_line.startswith('>    '):
                self._current_line = self._current_line.replace('> ', '')
        elif block is md4c.BlockType.TH:
            if self._inside_quoteblock:
                if not self._current_line.replace(' ', '') == '>':
                    self._current_line = removesuffix(self._current_line, '> ')
            self._current_line += '| '
            self._current_thead_aligns.append(details['align'].value)
        elif block is md4c.BlockType.TD:
            if self._inside_quoteblock:
                if not self._current_line.replace(' ', '') == '>':
                    self._current_line = removesuffix(self._current_line, '> ')
            self._current_line += '| '
        elif block is md4c.BlockType.QUOTE:
            if self._inside_liblock:
                self._save_current_msgid()
                self._save_current_line()
            self._inside_quoteblock = True
        elif block is md4c.BlockType.TABLE:
            self._inside_table = True
            if self._current_list_type and not self._inside_quoteblock:
                self._save_current_line()
        elif block is md4c.BlockType.HTML:
            self._inside_htmlblock = True

    def leave_block(self, block, details):
        # raise 'leave_block' event
        if raise_skip_event(
            self.events,
            'leave_block',
            self,
            block,
            details,
        ):
            return

        if block is md4c.BlockType.P:
            self._save_current_msgid()

            if self._inside_liblock:
                if self._inside_quoteblock:
                    indent = '   ' * len(self._current_list_type)
                    self._current_line = f'{indent}{self._current_line}'
                    self._save_current_line()
                else:
                    if self._inside_liblock_first_p:
                        self._inside_liblock_first_p = False
                    else:
                        indent = '   ' * len(self._current_list_type)
                        self._current_line = f'\n{indent}{self._current_line}'
                        self._save_current_line()
            else:
                self._save_current_line()

            self._inside_pblock = False
            if self._inside_quoteblock:
                self._current_line = '>'
                self._save_current_line()

        elif block is md4c.BlockType.CODE:
            self._save_current_msgid()
            self._inside_codeblock = False

            indent = ''
            if self._inside_liblock:
                indent += '   ' * len(self._current_list_type)
            self._current_line = self._current_line.rstrip('\n')
            self._save_current_line()
            if not self._inside_indented_codeblock:
                fence_chars = details['fence_char']*3
                self._current_line += f'{indent}{fence_chars}'

            self._save_current_line()
            if not self._inside_liblock:
                # prevent two newlines after indented code block
                if not self._inside_indented_codeblock:
                    self._save_current_line()
            self._inside_indented_codeblock = False
        elif block is md4c.BlockType.H:
            self._save_current_msgid()
            if not self._inside_quoteblock:
                self._current_line += '\n'
            else:
                self._current_line += '\n> '
            self._save_current_line()
            if self._inside_quoteblock:
                self._current_line += '> '
            self._inside_hblock = False
        elif block is md4c.BlockType.LI:
            self._save_current_msgid()
            self._inside_liblock = False
            if self._current_line:
                self._save_current_line()
        elif block is md4c.BlockType.UL:
            self._ul_marks.pop()
            self._current_list_type.pop()
            if self._inside_quoteblock:
                self._current_line += '> '
            if not self._ul_marks and self._outputlines[-1]:
                self._save_current_line()
        elif block is md4c.BlockType.OL:
            self._ol_marks.pop()
            self._current_list_type.pop()
            if self._inside_quoteblock:
                self._current_line += '> '
            if not self._ol_marks and self._outputlines[-1]:
                self._save_current_line()
        elif block is md4c.BlockType.TH or block is md4c.BlockType.TD:
            self._save_current_msgid()
            self._current_line += ' '
        elif block is md4c.BlockType.TR:
            if not self._current_thead_aligns:
                self._current_line += '|'
                self._save_current_line()
        elif block is md4c.BlockType.THEAD:
            import re

            # build thead separator
            thead_separator = ''
            if self._inside_quoteblock:
                _thead_split = re.split(r'[^\\](\|)', self._current_line)
                _thead_split = []
                for i, value in enumerate(self._current_line.split('|')):
                    _thead_split.extend([value, '|'])
                _thead_split.pop()
                if self._current_list_type:
                    _thead_split = _thead_split[1:]
                self._current_line += '|'
                thead_separator += '> '
            else:
                self._current_line += '|'
                _thead_split = re.split(r'[^\\](\|)', self._current_line)
                if self._current_list_type:
                    _thead_split.pop(0)
                    _thead_split.pop()
            thead_separator += '| '

            _antepenultimate_thead_i = len(_thead_split) - 2
            for i, title in enumerate(_thead_split):
                if (i % 2) != 0 or i > _antepenultimate_thead_i:
                    continue
                align = self._current_thead_aligns.pop(0)
                sep_left = '-' if align in [0, 3] else ':'
                sep_right = '-' if align in [0, 1] else ':'
                thead_separator += f'{sep_left}-{sep_right}'
                thead_separator += ' |'
                if i < len(_thead_split) - 3:
                    thead_separator += ' '

            indent = '   ' * len(self._current_list_type)
            self._current_line += f'\n{indent}{thead_separator}'
            self._save_current_line()
        elif block is md4c.BlockType.QUOTE:
            if self._outputlines[-1] == '>':
                self._outputlines.pop()
            if not self._inside_liblock:
                self._save_current_line()
            self._inside_quoteblock = False
        elif block is md4c.BlockType.TABLE:
            if not self._inside_quoteblock and not self._current_list_type:
                self._save_current_line()
            self._inside_table = False
        elif block is md4c.BlockType.HTML:
            self._inside_htmlblock = False

    def enter_span(self, span, details):
        # raise 'enter_span' event
        if raise_skip_event(
            self.events,
            'enter_span',
            self,
            span,
            details,
        ):
            return

        if self._inside_aspan:  # span inside link text
            try:
                self._current_aspan_text += self._enterspan_replacer[
                    span.value
                ]
            except KeyError:
                pass
        else:
            try:
                self._current_msgid += self._enterspan_replacer[span.value]
            except KeyError:
                pass

        if span is md4c.SpanType.A:
            self._inside_aspan = True

            if self._link_references is None:
                self._link_references = parse_link_references(self.content)

            self._current_aspan_href = details['href'][0][1]
            self._current_aspan_ref_target = None

            if details['title']:
                current_aspan_title = details['title'][0][1]
                for target, href, title in self._link_references:
                    if (
                        href == self._current_aspan_href and
                        title == current_aspan_title
                    ):
                        self._current_aspan_ref_target = target
                        break
            else:
                for target, href, title in self._link_references:
                    if href == self._current_aspan_href:
                        self._current_aspan_ref_target = target
                        break

        elif span is md4c.SpanType.CODE:
            self._inside_codespan = True
            self._codespan_start_index = len(self._current_msgid)-1
            self._codespan_inside_current_msgid = True
        elif span is md4c.SpanType.IMG:
            if self._link_references is None:
                self._link_references = parse_link_references(self.content)

            self._current_imgspan['title'] = '' if not details['title'] \
                else details['title'][0][1]
            self._current_imgspan['src'] = details['src'][0][1]
            self._current_imgspan['text'] = ''
        elif span is md4c.SpanType.WIKILINK:
            self._current_wikilink_target = details['target'][0][1]

    def leave_span(self, span, details):
        # raise 'leave_span' event
        if raise_skip_event(
            self.events,
            'leave_span',
            self,
            span,
            details,
        ):
            return

        if span is md4c.SpanType.WIKILINK:
            self._current_msgid += polib.escape(self._current_wikilink_target)
            self._current_wikilink_target = None

        if self._inside_aspan:  # span inside link text
            try:
                self._current_aspan_text += self._leavespan_replacer[
                    span.value
                ]
            except KeyError:
                pass
        else:
            try:
                self._current_msgid += self._leavespan_replacer[span.value]
            except KeyError:
                pass

        if span is md4c.SpanType.A:
            if self._current_aspan_ref_target:  # referenced link
                self._current_msgid += (
                    f'[{self._current_aspan_text}]'
                    f'[{self._current_aspan_ref_target}]'
                )
                self._current_aspan_ref_target = None
            else:
                if self._current_aspan_text == self._current_aspan_href:
                    # autolink vs link clash (see implementation notes)
                    self._current_msgid += f'<{self._current_aspan_text}'
                    if details['title']:
                        escaped_title = polib.escape(details['title'][0][1])
                        self._current_msgid += f' "{escaped_title}"'
                    self._current_msgid += '>'
                elif self._current_aspan_href:
                    self._current_msgid += (
                        f'[{self._current_aspan_text}]'
                        f'({self._current_aspan_href}'
                    )
                    if details['title']:
                        self._aimg_title_inside_current_msgid = True
                        escaped_title = polib.escape(details['title'][0][1])
                        self._current_msgid += f' "{escaped_title}"'
                    self._current_msgid += ')'
            self._current_aspan_href = None
            self._inside_aspan = False
            self._current_aspan_text = ''
        elif span is md4c.SpanType.CODE:
            self._inside_codespan = False
            self._current_msgid += (
                self._codespan_backticks * self.code_end_string
            )
            self._codespan_backticks = None
        elif span is md4c.SpanType.IMG:
            referenced_target, imgspan_title = (None, None)
            imgspan_src = details['src'][0][1]
            if details['title']:
                imgspan_title = details['title'][0][1]
                for target, href, title in self._link_references:
                    if href == imgspan_src and title == imgspan_title:
                        referenced_target = target
                        break
            else:
                for target, href, _ in self._link_references:
                    if href == imgspan_src:
                        referenced_target = target
                        break

            alt_text = self._current_imgspan['text']
            img_markup = f'![{alt_text}]'
            if referenced_target:
                img_markup += f'[{referenced_target}]'
            else:
                img_markup += f'({imgspan_src}'
                if imgspan_title:
                    img_markup += f' "{imgspan_title}"'
                img_markup += ')'

            self._current_imgspan = {}

            if self._inside_aspan:
                self._current_aspan_text += img_markup
            else:
                self._current_msgid += img_markup

    def text(self, block, text):
        # raise 'text' event
        if raise_skip_event(
            self.events,
            'text',
            self,
            block,
            text,
        ):
            return

        if not self._inside_htmlblock:
            if not self._inside_codeblock:
                if self._inside_liblock and text == '\n':
                    text = ' '
                if self._current_imgspan:
                    self._current_imgspan['text'] = text
                    return
                elif self._inside_codespan:
                    self._codespan_backticks = min_not_max_chars_in_a_row(
                        self.code_start_string,
                        text,
                    ) - 1
                    self._current_msgid = (
                        f'{self._current_msgid[:self._codespan_start_index]}'
                        f'{self._codespan_backticks * self.code_start_string}'
                        f'{self._current_msgid[self._codespan_start_index:]}'
                    )
                    if self._inside_aspan:
                        self._current_aspan_text += text
                        return
                elif self._inside_aspan:
                    self._current_aspan_text += text
                    return
                elif text == self.italic_start_string:
                    text = self.italic_start_string_escaped
                elif text == self.code_start_string:
                    text = self.code_start_string_escaped
                elif text == self.code_end_string:  # pragma: no cover
                    text = self.code_end_string_escaped
                elif text == self.italic_end_string:  # pragma: no cover
                    text = self.italic_end_string_escaped

                if self._inside_pblock:
                    text = text.replace('\n', ' ')
                if text == self._current_aspan_href:
                    # self-referenced wiki or inline link
                    self._current_aspan_href = None

                elif self._current_wikilink_target:
                    if text != self._current_wikilink_target:
                        self._current_wikilink_target = (
                            f'{self._current_wikilink_target}|{text}'
                        )
                    return
                self._current_msgid += text
            else:
                if self._inside_liblock:
                    indent = '   ' * len(self._current_list_type)
                    if self._current_line[:len(indent)+1] != indent:
                        self._current_line += indent
                self._current_msgid += text
        else:
            self._process_command(text)

    def _append_link_references(self):
        if self._link_references:
            self._disable_next_line = False
            self._disable = False

            # 'link_reference' event
            pre_events = self.events.get('link_reference')

            added_references = []  # don't repeat references
            for target, href, title in self._link_references:
                if pre_events:
                    skip = False
                    for event in pre_events:
                        if event(self, target, href, title) is False:
                            skip = True
                    if skip:
                        continue
                title_part = f' "{title}"' if title else ''
                href_title = f' {href}{title_part}'
                if href_title in added_references:
                    continue

                msgid = f'[{target}]:{href_title}'
                self._outputlines.append(
                    self._translate_msgid(msgid, None, None),
                )
                added_references.append(href_title)
            self._outputlines.append('')

    def translate(
        self,
        filepath_or_content,
        save=None,
        md_encoding='utf-8',
    ):
        self.content = to_file_content_if_is_file(
            filepath_or_content,
            encoding=md_encoding,
        )

        self.translations, self.translations_with_msgctxt = (
            pofiles_to_unique_translations_dicts(self.pofiles)
        )

        parser = md4c.GenericParser(
            0,
            **{ext: True for ext in self.extensions},
        )
        parser.parse(
            self.content,
            self.enter_block,
            self.leave_block,
            self.enter_span,
            self.leave_span,
            self.text,
        )
        self._append_link_references()  # add link references to the end

        self._disable_next_line = False
        self._disable = False
        self._enable_next_line = False
        self._link_references = None

        self.output = '\n'.join(self._outputlines)

        if save:
            if self._saved_files_changed is False:
                self._saved_files_changed = save_file_checking_file_changed(
                    save,
                    self.output,
                    encoding=md_encoding,
                )
            else:
                with open(save, 'w', encoding=md_encoding) as f:
                    f.write(self.output)
        return self.output


def pofile_to_markdown(
    filepath_or_content,
    pofiles,
    ignore=[],
    save=None,
    md_encoding='utf-8',
    po_encoding=None,
    command_aliases={},
    wrapwidth=80,
    events={},
    debug=False,
    **kwargs,
):
    r"""Translate Markdown content or file using PO files as reference.

    This implementation reproduces the same valid Markdown output, given the
    provided AST, with replaced translations, but doesn't rebuilds the same
    input format as Markdown is just a subset of HTML.

    Args:
        filepath_or_content (str): Markdown filepath or content to translate.
        pofiles (str, list) Glob or list of globs matching a set of PO files
            from where to extract messages to make the replacements translating
            strings.
        ignore (list): Paths of PO files to ignore. Useful when a glob does not
            fit your requirements indicating the files to extract content.
            Also, filename or a dirname can be defined without indicate the
            full path.
        save (str): Saves the output content in file whose path is specified
            at this parameter.
        md_encoding (str): Markdown content encoding.
        po_encoding (str): PO files encoding. If you need different encodings
            for each file, you must define it in the "Content-Type" field of
            each PO file metadata, in the form
            ``"Content-Type: text/plain; charset=<ENCODING>\n"``.
        command_aliases (dict): Mapping of aliases to use custom mdpo command
            names in comments. The ``mdpo-`` prefix in command names resolution
            is optional. For example, if you want to use ``<!-- mdpo-on -->``
            instead of ``<!-- mdpo-enable -->``, you can pass the dictionaries
            ``{"mdpo-on": "mdpo-enable"}`` or ``{"mdpo-on": "enable"}`` to this
            parameter.
        wrapwidth (int): Maximum width used rendering the Markdown output.
        events (dict): Preprocessing events executed during the translation
            process. You can use these to customize the output. Takes functions
            are values. If one of these functions returns ``False``, that part
            of the translation process is skipped by po2md. The available
            events are:

            * ``enter_block(self, block, details)``: Executed when the parsing
              a Markdown block starts.
            * ``leave_block(self, block, details)``: Executed when the parsing
              a Markdown block ends.
            * ``enter_span(self, span, details)``: Executed when the parsing of
              a Markdown span starts.
            * ``leave_span(self, span, details)``: Executed when the parsing of
              a Markdown span ends.
            * ``text(self, block, text)``: Executed when the parsing of text
              starts.
            * ``command(self, mdpo_command, comment, original command)``:
              Executed when a mdpo HTML command is found.
            * ``msgid(self, msgid, msgstr, msgctxt, tcomment, flags)``:
              Executed when a msgid is going to be replaced.
            * ``link_reference(self, target, href, title)``: Executed when each
              reference link is being written in the output (at the end of the
              translation process).
        debug (bool): Add events displaying all parsed elements in the
            translation process.

    Returns:
        str: Markdown output file with translated content.
    """
    return Po2Md(
        pofiles,
        ignore=ignore,
        po_encoding=po_encoding,
        command_aliases=command_aliases,
        wrapwidth=wrapwidth,
        events=events,
        debug=debug,
        **kwargs,
    ).translate(
        filepath_or_content,
        save=save,
        md_encoding=md_encoding,
    )



#!/usr/bin/env python

"""md2po command line interface.

See :ref:`md2po CLI<cli:md2po>`.
"""

import argparse
import sys

from mdpo.cli import (
    CLOSE_QUOTE_CHAR,
    OPEN_QUOTE_CHAR,
    add_command_alias_argument,
    add_common_cli_first_arguments,
    add_debug_option,
    add_encoding_arguments,
    add_extensions_argument,
    add_nolocation_option,
    add_pre_commit_option,
    cli_codespan,
    parse_command_aliases_cli_arguments,
    parse_metadata_cli_arguments,
)
from mdpo.context import environ
from mdpo.md2po import Md2Po
from mdpo.md4c import DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS


DESCRIPTION = (
    'Utility like xgettext to extract Markdown contents dumping them'
    ' inside PO files.'
)


def build_parser():
    parser = argparse.ArgumentParser(description=DESCRIPTION, add_help=False)
    add_common_cli_first_arguments(parser)
    parser.add_argument(
        'files_or_content', metavar='GLOBS_FILES_OR_CONTENT',
        nargs='*',
        help='Globs to markdown input files, paths to files or Markdown'
             ' content. If not provided, will be read from STDIN.',
    )
    parser.add_argument(
        '-i', '--ignore', dest='ignore', default=[], action='append',
        help='Path to a file to ignore. This argument can be passed multiple'
             ' times.',
        metavar='PATH',
    )
    parser.add_argument(
        '-po', '--po-filepath', '--pofilepath', dest='po_filepath',
        default=None,
        help='Merge new msgids in the po file indicated at this parameter (if'
             f' {cli_codespan("--save")} argument is passed) or use the msgids'
             ' of the file as reference for mark not found as obsoletes if'
             f' {cli_codespan("--merge-pofiles")} parameter is not passed.',
        metavar='OUTPUT_PO_FILEPATH',
    )
    parser.add_argument(
        '-s', '--save', dest='save', action='store_true',
        help='Save new found msgids to the po file indicated as parameter'
             f' {cli_codespan("--po-filepath")}. Passing this option without'
             f' defining the argument {cli_codespan("--po-filepath")} will'
              ' raise an error.',
    )
    parser.add_argument(
        '-mo', '--mo-filepath', '--mofilepath', dest='mo_filepath',
        default=None,
        help='The resulting PO file will be compiled to a mofile and saved in'
             ' the path specified at this parameter.',
        metavar='OUTPUT_MO_FILEPATH',
    )
    parser.add_argument(
        '-p', '--plaintext', dest='plaintext',
        action='store_true',
        help='Do not include markdown markup characters in extracted msgids'
             f' for {cli_codespan("**bold text**", cli=False)},'
             f' {cli_codespan("*italic text*", cli=False)},'
             f' {cli_codespan("``inline code``")} and'
             f' {cli_codespan("[link](target)")}.',
    )
    parser.add_argument(
        '-w', '--wrapwidth', dest='wrapwidth', metavar='N/inf', type=str,
        default='78',
        help='Wrap width for po file indicated at'
             f' {cli_codespan("--po-filepath")} parameter. If negative,'
             ' \'0\' or \'inf\', the PO file content will not be wrapped.',
    )
    parser.add_argument(
        '-m', '--merge-po-files', '--merge-pofiles',
        dest='mark_not_found_as_obsolete',
        action='store_false',
        help='Messages not found which are already stored in the PO file'
             f' passed as {cli_codespan("--po-filepath")} argument will not be'
             ' marked as obsolete.',
    )
    parser.add_argument(
        '-r', '--remove-not-found',
        dest='preserve_not_found',
        action='store_false',
        help='Messages not found which are already stored in the PO file'
             f' passed as {cli_codespan("--po-filepath")} parameter will be'
             ' removed. Only has effect used in combination with'
             f' {cli_codespan("--merge-pofiles")}.',
    )
    add_nolocation_option(parser)
    add_extensions_argument(parser)
    add_encoding_arguments(
        parser,
        po_encoding_help='Resulting PO file encoding.',
    )
    parser.add_argument(
        '-a', '--xheaders', dest='xheaders',
        action='store_true',
        help='Include mdpo specification x-headers. These only will be'
             ' included if you do not pass the parameter'
             f' {cli_codespan("--plaintext")}.',
    )
    parser.add_argument(
        '-c', '--include-codeblocks',
        dest='include_codeblocks', action='store_true',
        help='Include all code blocks found inside PO file result. This is'
             ' useful if you want to translate all your blocks of code.'
             ' Equivalent to append'
             f' {cli_codespan("<!-- mdpo-include-codeblock -->")} command'
             ' before each code block.',
    )
    parser.add_argument(
        '--ignore-msgids', dest='ignore_msgids', default=None,
        help='Path to a plain text file where all msgids to ignore from being'
             ' extracted are located, separated by newlines.',
    )

    # patch for sphinx-argparse-cli compatibility (use Unicode quotation marks)
    example_codespan = (
        f'-d {OPEN_QUOTE_CHAR}Content-Type: text/plain;'
        f' charset=utf-8{CLOSE_QUOTE_CHAR}'
        f' -d {OPEN_QUOTE_CHAR}Language: es{CLOSE_QUOTE_CHAR}'
    )
    metadata_help_example = (
        ' For example, to define UTF-8 encoding and Spanish language use'
        f' {cli_codespan(example_codespan)}.'
    )
    parser.add_argument(
        '-d', '--metadata', dest='metadata', default=[], action='append',
        metavar='Key:Value',
        help='Custom metadata key-value pairs to include in the produced'
             ' PO file. This argument can be passed multiple times.'
             ' If the file contains previous metadata fields, these will'
             ' be updated preserving the values of the already defined.'
             f'{metadata_help_example}',
    )

    add_command_alias_argument(parser)
    add_debug_option(parser)
    add_pre_commit_option(parser)
    return parser


def parse_options(args=[]):
    parser = build_parser()
    if '-h' in args or '--help' in args:
        parser.print_help()
        sys.exit(1)
    opts, unknown = parser.parse_known_args(args)

    if not sys.stdin.isatty():
        files_or_content = sys.stdin.read().strip('\n')
    if isinstance(opts.files_or_content, list) and opts.files_or_content:
        if len(opts.files_or_content) == 1:
            files_or_content = opts.files_or_content[0]
        else:
            files_or_content = opts.files_or_content
    opts.files_or_content = files_or_content

    if opts.extensions is None:
        opts.extensions = DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS

    if opts.ignore_msgids is not None:
        with open(opts.ignore_msgids) as f:
            opts.ignore_msgids = f.read().splitlines()
    else:
        opts.ignore_msgids = []

    opts.command_aliases = parse_command_aliases_cli_arguments(
        opts.command_aliases,
    )
    opts.metadata = parse_metadata_cli_arguments(
        opts.metadata,
    )

    return opts


def run(args=[]):
    with environ(_MDPO_RUNNING='true'):
        opts = parse_options(args)

        init_kwargs = dict(
            ignore=opts.ignore,
            plaintext=opts.plaintext,
            mark_not_found_as_obsolete=opts.mark_not_found_as_obsolete,
            preserve_not_found=opts.preserve_not_found,
            location=opts.location,
            extensions=opts.extensions,
            xheaders=opts.xheaders,
            include_codeblocks=opts.include_codeblocks,
            ignore_msgids=opts.ignore_msgids,
            command_aliases=opts.command_aliases,
            metadata=opts.metadata,
            debug=opts.debug,
            _check_saved_files_changed=opts.check_saved_files_changed,
        )

        extract_kwargs = dict(
            po_filepath=opts.po_filepath,
            save=opts.save,
            mo_filepath=opts.mo_filepath,
            po_encoding=opts.po_encoding,
            md_encoding=opts.md_encoding,
            wrapwidth=opts.wrapwidth,
        )

        md2po = Md2Po(opts.files_or_content, **init_kwargs)
        pofile = md2po.extract(**extract_kwargs)

        if not opts.quiet:
            sys.stdout.write(f'{pofile.__unicode__()}\n')

        # pre-commit mode
        if opts.check_saved_files_changed and md2po._saved_files_changed:
            return (pofile, 1)

    return (pofile, 0)


def main():
    sys.exit(run(args=sys.argv[1:])[1])  # pragma: no cover


if __name__ == '__main__':
    main()



"""I/O mdpo utitlites tests."""

import glob
import html
import os
import tempfile
import uuid

import pytest

from mdpo.io import (
    filter_paths,
    save_file_checking_file_changed,
    to_file_content_if_is_file,
    to_files_or_content,
)


EMPTY_FILES_DIRNAME = 'empty-files'
EMPTY_FILES_GLOBSTR = os.path.join('test', EMPTY_FILES_DIRNAME, '**', '**.md')
EMPTY_FILES_GLOB = glob.glob(EMPTY_FILES_GLOBSTR)

MD_CONTENT_EXAMPLE = "# Hello\n\nI'm markdown\n"


class TestFilterPaths:
    def empty_file_path(self, directory, filename=None):
        args = () if not filename else (filename,)
        return os.path.join('test', EMPTY_FILES_DIRNAME, directory, *args)

    def test_ignore_files_by_filename(self):
        filepaths = filter_paths(
            EMPTY_FILES_GLOB,
            ignore_paths=['foo04.md', 'bar02.md'],
        )
        assert filepaths == [
            self.empty_file_path('bar', 'bar01.md'),
            self.empty_file_path('bar', 'bar03.md'),
            self.empty_file_path('foo', 'foo01.md'),
            self.empty_file_path('foo', 'foo02.md'),
            self.empty_file_path('foo', 'foo03.md'),
        ]

    def test_ignore_directory_by_dirname(self):
        filepaths = filter_paths(EMPTY_FILES_GLOB, ignore_paths=['foo'])

        assert filepaths == [
            self.empty_file_path('bar', 'bar01.md'),
            self.empty_file_path('bar', 'bar02.md'),
            self.empty_file_path('bar', 'bar03.md'),
        ]

    def test_ignore_files_by_filepath(self):
        filepaths = filter_paths(
            EMPTY_FILES_GLOB,
            ignore_paths=[
                self.empty_file_path('foo', 'foo04.md'),
                self.empty_file_path('bar', 'bar02.md'),
            ],
        )

        assert filepaths == [
            self.empty_file_path('bar', 'bar01.md'),
            self.empty_file_path('bar', 'bar03.md'),
            self.empty_file_path('foo', 'foo01.md'),
            self.empty_file_path('foo', 'foo02.md'),
            self.empty_file_path('foo', 'foo03.md'),
        ]

    def test_ignore_files_by_dirpath(self):
        filepaths = filter_paths(
            EMPTY_FILES_GLOB,
            ignore_paths=[self.empty_file_path('foo')],
        )

        assert filepaths == [
            self.empty_file_path('bar', 'bar01.md'),
            self.empty_file_path('bar', 'bar02.md'),
            self.empty_file_path('bar', 'bar03.md'),
        ]


class TestToGlobOrContent:
    def test_glob(self):
        is_glob, parsed = to_files_or_content(EMPTY_FILES_GLOBSTR)
        assert is_glob
        assert parsed == EMPTY_FILES_GLOB

    def test_content(self):
        is_glob, parsed = to_files_or_content(MD_CONTENT_EXAMPLE)
        assert not is_glob
        assert parsed == MD_CONTENT_EXAMPLE

    @pytest.mark.parametrize('argtype', (tuple, list, set, iter))
    def test_list(self, tmp_file, argtype):
        with tmp_file('foo\n', '.md') as foo_path, \
                tmp_file('bar\n', '.md') as bar_path:
            value = argtype({foo_path, bar_path})
            is_glob, parsed = to_files_or_content(value)
            assert is_glob
            assert parsed == value

    def test_bad_glob_characters_range(self):
        content = html.escape('[s-m]')
        is_glob, parsed = to_files_or_content(content)
        assert not is_glob
        assert parsed == content


class TestToFileContentIfIsFile:
    def test_file(self):
        with tempfile.NamedTemporaryFile('w+') as tmpfile:
            tmpfile.write(MD_CONTENT_EXAMPLE)
            tmpfile.seek(0)
            md_content = to_file_content_if_is_file(tmpfile.name)
        assert md_content == MD_CONTENT_EXAMPLE

    def test_content(self):
        md_content = to_file_content_if_is_file(MD_CONTENT_EXAMPLE)
        assert md_content == MD_CONTENT_EXAMPLE


def test_save_file_checking_file_changed(tmp_file):
    tempfile_path = os.path.join(
        tempfile.gettempdir(),
        f'mdpo--{uuid.uuid4().hex[:8]}',
    )

    changed = save_file_checking_file_changed(tempfile_path, 'foo\n')
    assert changed

    if os.path.isfile(tempfile_path):
        os.remove(tempfile_path)



import os


def test_devref_index():
    devref_index_path = os.path.join('docs', 'devref', 'index.rst')
    with open(devref_index_path) as f:
        devref_index_content = f.read()

    utils_index, implementations_index = [], []
    for line in devref_index_content.splitlines():
        if line.startswith('   mdpo'):
            if line.endswith('.index'):
                implementations_index.append(
                    '/'.join(line.lstrip(' ').split('/')[:2]),
                )
            else:
                utils_index.append(line.lstrip(' ').replace('/mdpo', ''))

    utils_modules, implementation_packages = [], []
    for filename in os.listdir('mdpo'):
        if filename.startswith('__'):  # ignore __init__.py
            continue

        filepath = os.path.join('mdpo', filename)
        if os.path.isfile(filepath):
            basename_no_ext = filename.rstrip('.py')
            utils_modules.append(f'mdpo.{basename_no_ext}')
        else:
            implementation_packages.append(filepath)

    # synchronize utilities with devref docs index
    for util_module in utils_modules:
        assert util_module in utils_index
    for util_doc in utils_index:
        assert util_doc in utils_modules

    # synchronize implementations with devref docs index
    for implementation_package in implementation_packages:
        assert implementation_package in implementations_index
    for implementation_doc in implementations_index:
        assert implementation_doc in implementation_packages



"""mdpo command line interface utilities."""

import argparse
import sys

from mdpo import __version__
from mdpo.md4c import DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS
from mdpo.text import and_join, parse_escaped_pairs


SPHINX_IS_RUNNING = 'sphinx' in sys.modules
OPEN_QUOTE_CHAR = '”' if SPHINX_IS_RUNNING else '"'
CLOSE_QUOTE_CHAR = '”' if SPHINX_IS_RUNNING else '"'


def cli_codespan(value, cli=True, sphinx=True):
    """Command line codespan wrapper.

    This is a compatibility function to make CLI codespans looks good in
    sphinx-argparse-cli documentation and using ``--help`` option in CLI
    itself. sphinx-argparse-cli expects the usage of double backticks for
    codespans, but that format is ugly in CLI.

    Args:
        value (str): Value to wrap.
        cli (bool): Wrap when used from command line.
        sphinx (bool): Wrap when used from Sphinx.
    """
    if SPHINX_IS_RUNNING:
        return f'``{value}``' if sphinx else value
    else:
        return f'\'{value}\'' if cli else value


def parse_escaped_pairs_cli_argument(
    pairs,
    value_error_message,
    key_error_message,
):
    """Parses a key argument made by key-value pairs.

    If an error happens, shows an appropiate message and exists with code 1.

    Args:
        pairs (list): List of key-value pairs.
        value_error_message (str): Error message schema shown when a pair
            can't be parsed.
        key_error_message (str): Error message schema shown when a key is
            repeated.

    Returns:
        dict: Parsed key-value pairs.
    """
    try:
        return parse_escaped_pairs(pairs)
    except ValueError as err:
        sys.stderr.write(value_error_message.format(err.args[0]))
        sys.exit(1)
    except KeyError as err:
        sys.stderr.write(key_error_message.format(err.args[0]))
        sys.exit(1)


def parse_command_aliases_cli_arguments(command_aliases):
    """Parse ``--command-alias`` argument values passed to CLIs.

    If a value can't be passed or a custom command is duplicated, writes an
    appropiate error message to STDERR and exits with code 1.

    Args:
        command_aliases (list): Values taken by ``--command-alias`` arguments.

    Returns:
        dict: Command aliases mapping ni the format accepted by the API.
    """
    return parse_escaped_pairs_cli_argument(
        command_aliases,
        (
            "The value '{}' passed to argument --command-alias"
            " can't be parsed. Please, separate the pair "
            "'<custom-command:mdpo-command>' with a ':' character.\n"
        ),
        (
            "Multiple resolutions for '{}' alias passed to"
            ' --command-alias arguments.\n'
        ),
    )


def parse_metadata_cli_arguments(metadata):
    """Parse ``--metadata`` argument values passed to CLIs.

    If a value can't be passed or a metadata key is duplicated, writes an
    appropiate error message to STDERR and exits with code 1.

    Args:
        metadata (list): Values taken by ``--metadata`` arguments.

    Returns:
        dict: Metadata mapping ni the format accepted by the API.
    """
    return parse_escaped_pairs_cli_argument(
        metadata,
        (
            "The value '{}' passed to argument --metadata"
            " can't be parsed. Please, separate the pair "
            "'<key:value>' with a ':' character.\n"
        ),
        (
            "Repeated key '{}' passed to --metadata arguments.\n"
        ),
    )


def add_common_cli_first_arguments(parser, quiet=True):
    """Add common mdpo arguments to an argument parser at the beginning.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Arguments parser to
            extend.
        quiet (bool): Include the argument ``-q/--quiet``.
    """
    parser.add_argument(
        '-h', '--help', action='help',
        help='Show this help message and exit.',
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        '-v', '--version', action='version',
        version=f'%(prog)s {__version__}',
        help='Show program version number and exit.',
    )
    if quiet:
        parser.add_argument(
            '-q', '--quiet', action='store_true',
            help='Do not print output to STDOUT.',
        )


def add_command_alias_argument(parser):
    """Add the ``--command-alias`` argument to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Arguments parser to
            extend.
    """
    mdpo_on_mdpo_enable = cli_codespan(
        f'--command-alias {OPEN_QUOTE_CHAR}mdpo-on:mdpo-enable'
        f'{CLOSE_QUOTE_CHAR}',
    )
    mdpo_on_enable = cli_codespan(
        f'--command-alias {OPEN_QUOTE_CHAR}mdpo-on:enable'
        f'{CLOSE_QUOTE_CHAR}',
    )
    command_alias_help_example = (
        ' For example, if you want to use "<!-- mdpo-on -->" instead of'
        f' "<!-- mdpo-enable -->", you can pass either {mdpo_on_mdpo_enable}'
        f' or {mdpo_on_enable} arguments.'
    )
    parser.add_argument(
        '--command-alias', dest='command_aliases', default=[], action='append',
        metavar='CUSTOM-COMMAND:MDPO-COMMAND',
        help='Aliases to use custom mdpo command names in comments. This'
             ' argument can be passed multiple times in the form'
             ' "<custom-command>:<mdpo-command>". The \'mdpo-\' prefix in'
             ' command names resolution is optional.'
             f'{command_alias_help_example}',
    )


def add_extensions_argument(parser):
    """Add the ``-x/--extension`` argument to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Arguments parser to
            extend.
    """
    parser.add_argument(
        '-x', '--extension', '--ext', dest='extensions', action='append',
        default=None,
        help='md4c extension used to parse markdown content formatted as'
             ' pymd4c extension keyword arguments. This argument can be passed'
             ' multiple times. If it is not passed, next extensions are used:'
             f' {and_join(DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS)}.'
             ' You can see all available at'
             ' https://pymd4c.dcpx.org/api.html#parser-option-flags',
        metavar='EXTENSION',
    )


def add_debug_option(parser):
    """Add the ``-D/--debug`` option to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Arguments parser to
            extend.
    """
    parser.add_argument(
        '-D', '--debug', dest='debug', action='store_true',
        help='Print useful messages in the parsing process showing the'
             ' contents of all Markdown elements.',
    )


def add_nolocation_option(parser):
    """Add the ``--no-location/--nolocation`` option to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Arguments parser to
            extend.
    """
    parser.add_argument(
        '--no-location', '--nolocation', dest='location', action='store_false',
        help='Do not write \'#: filename:line\' lines. Note that using this'
             ' option makes it harder for technically skilled translators to'
             ' understand the context of each message. Same as'
             f' {cli_codespan("gettext --no-location")}.',
    )


def add_encoding_arguments(
        parser,
        po_encoding_help=None,
        markup_encoding='md',
):
    """Add ``--po-encoding`` ``--md_encoding`` arguments to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Arguments parser to
            extend.
        po_encoding_help (str): Help text for the ``--po-encoding`` argument.
        markup_encoding (str): Type of markup encoding, can be ``"md"`` or
            ``"html"``.
    """
    parser.add_argument(
        f'--{markup_encoding}-encoding',
        dest=f'{markup_encoding}_encoding',
        default='utf-8',
        help=f'{"Markdown" if markup_encoding == "md" else "HTML"}'
             ' content encoding.',
        metavar='ENCODING',
    )

    po_encoding_help = (
        'PO files encoding. If you need different encodings for each'
        ' file, you must define them in the "Content-Type" field of each'
        ' PO file metadata, in the form \'Content-Type: text/plain;'
        ' charset=<ENCODING>\'.'
    ) if po_encoding_help is None else po_encoding_help
    parser.add_argument(
        '--po-encoding', dest='po_encoding', default=None, metavar='ENCODING',
        help=po_encoding_help,
    )


def add_pre_commit_option(parser):
    """Add the ``--pre-commit`` option to an argument parser.

    Args:
        parser (:py:class:`argparse.ArgumentParser`): Arguments parser to
            extend.
    """
    parser.add_argument(
        '--pre-commit', dest='check_saved_files_changed', action='store_true',
        help='Run in pre-commit mode, which returns code 1 at exit when a file'
             ' has been changed or previously did not exist.',
    )



"""mdpo package."""

__description__ = ('Markdown files translation using PO files.')
__title__ = 'mdpo'
__version__ = '0.3.76'
__all__ = [
    '__description__',
    '__title__',
    '__version__',
]


def __getattr__(name):
    """Implements PEP 562 to avoid uneeded imports in CLIs."""
    import importlib
    func_package_map = {
        'markdown_pofile_to_html': 'mdpo2html',
        'markdown_to_pofile': 'md2po',
        'markdown_to_pofile_to_markdown': 'md2po2md',
        'pofile_to_markdown': 'po2md',
    }
    try:
        return getattr(
            importlib.import_module(f'mdpo.{func_package_map[name]}'),
            name,
        )
    except KeyError:
        raise ImportError(
            f'cannot import name \'{name}\' from \'mdpo\' ({__file__})',
            name=name,
            path='mdpo',
        ) from None


def __dir__():
    return __all__ + [
        'markdown_pofile_to_html',
        'markdown_to_pofile',
        'markdown_to_pofile_to_markdown',
        'pofile_to_markdown',
    ]



#!/usr/bin/env python

"""mdpo2html command line interface.

See :ref:`mdpo2html CLI<cli:mdpo2html>`.
"""

import argparse
import itertools
import sys

from mdpo.cli import (
    add_command_alias_argument,
    add_common_cli_first_arguments,
    add_encoding_arguments,
    add_pre_commit_option,
    cli_codespan,
    parse_command_aliases_cli_arguments,
)
from mdpo.context import environ
from mdpo.mdpo2html import MdPo2HTML


DESCRIPTION = 'HTML-produced-from-Markdown file translator using PO files.'


def build_parser():
    parser = argparse.ArgumentParser(description=DESCRIPTION, add_help=False)
    add_common_cli_first_arguments(parser)
    parser.add_argument(
        'filepath_or_content', metavar='FILEPATH_OR_CONTENT',
        nargs='*',
        help='HTML file path or content to translate. If not provided, will be'
             ' read from STDIN.',
    )
    parser.add_argument(
        '-p', '--po-files', '--pofiles', metavar='POFILES', action='append',
        nargs='*', dest='pofiles',
        help='Glob matching a set of PO files from where to extract references'
             ' to make the replacements translating strings. This argument'
             ' can be passed multiple times.',
    )
    parser.add_argument(
        '-i', '--ignore', dest='ignore', default=[], action='append',
        help=f'Filepaths to ignore when {cli_codespan("--pofiles")} argument'
             ' value is a glob. This argument can be passed multiple times.',
        metavar='PATH',
    )
    parser.add_argument(
        '-s', '--save', dest='save', default=None,
        help='Saves the output content in file whose path is specified at this'
             ' parameter.', metavar='PATH',
    )
    add_encoding_arguments(parser, markup_encoding='html')
    add_command_alias_argument(parser)
    add_pre_commit_option(parser)
    return parser


def parse_options(args):
    parser = build_parser()
    if '-h' in args or '--help' in args:
        parser.print_help()
        sys.exit(1)
    opts = parser.parse_args(args)

    filepath_or_content = ''
    if not sys.stdin.isatty():
        filepath_or_content += sys.stdin.read().strip('\n')
    if (
        isinstance(opts.filepath_or_content, list)
        and opts.filepath_or_content
    ):
        filepath_or_content += opts.filepath_or_content[0]
    opts.filepath_or_content = filepath_or_content

    opts.command_aliases = parse_command_aliases_cli_arguments(
        opts.command_aliases,
    )

    opts.pofiles = set(itertools.chain(*opts.pofiles))  # flatten

    return opts


def run(args=[]):
    with environ(_MDPO_RUNNING='true'):
        opts = parse_options(args)

        mdpo2html = MdPo2HTML(
            opts.pofiles,
            ignore=opts.ignore,
            po_encoding=opts.po_encoding,
            command_aliases=opts.command_aliases,
            _check_saved_files_changed=opts.check_saved_files_changed,
        )
        output = mdpo2html.translate(
            opts.filepath_or_content,
            save=opts.save,
            html_encoding=opts.html_encoding,
        )

        if not opts.quiet and not opts.save:
            sys.stdout.write(f'{output}\n')

        if opts.check_saved_files_changed and mdpo2html._saved_files_changed:
            return (output, 1)

    return (output, 0)


def main():
    sys.exit(run(args=sys.argv[1:])[1])  # pragma: no cover


if __name__ == '__main__':
    main()



"""HTML mdpo command utitlites tests."""

import pytest

from mdpo.command import (
    get_valid_mdpo_command_names,
    normalize_mdpo_command,
    normalize_mdpo_command_aliases,
    parse_mdpo_html_command,
)
from mdpo.text import removeprefix


@pytest.mark.parametrize(
    ('value', 'expected_command', 'expected_comment'), (
        ('<!-- mdpo-include-codeblock -->', 'mdpo-include-codeblock', None),
        ('<!-- mdpo_include_codeblock -->', 'mdpo_include_codeblock', None),
        ('<!-- -->', None, None),

        # comments include spaces at the end, implementations must rstrip them
        ('<!-- mdpo-include Hello world -->', 'mdpo-include', 'Hello world '),
        ('<!--mdpo-include Hello world-->', 'mdpo-include', 'Hello world'),
        ('<!-- mdpo-include Hello world-->', 'mdpo-include', 'Hello world'),
        (
            '<!-- mdpo-include Hello world\n-->',
            'mdpo-include',
            'Hello world\n',
        ),
        (
            '<!-- mdpo-include Hello world\n -->',
            'mdpo-include',
            'Hello world\n ',
        ),
        ('<!--mdpo-includeHello world-->', 'mdpo-includeHello', 'world'),
        ('<!--mdpo_include Hello world-->', 'mdpo_include', 'Hello world'),
    ),
)
def test_parse_mdpo_html_command(value, expected_command, expected_comment):
    command, comment = parse_mdpo_html_command(value)
    assert command == expected_command
    assert comment == expected_comment


@pytest.mark.parametrize(
    ('value', 'expected_command'), (
        ('include-codeblock', 'mdpo-include-codeblock'),
        ('include_codeblock', None),
        ('include', 'mdpo-include'),
        ('mdpo-include', 'mdpo-include'),
        ('mdpo-mdpo-include', None),
    ),
)
def test_normalize_mdpo_command(value, expected_command):
    assert normalize_mdpo_command(value) == expected_command


@pytest.mark.parametrize(
    ('command_aliases', 'expected_result'), (
        (
            {'inclusion': 'mdpo-include'},
            {'inclusion': 'mdpo-include'},
        ),
        (
            {'mdpo-inclusion': 'mdpo-include'},
            {'mdpo-inclusion': 'mdpo-include'},
        ),
        (
            {'inclusion': 'include'},
            {'inclusion': 'mdpo-include'},
        ),
        (
            {'inclusion': 'inclusion'},
            ValueError,
        ),
    ),
)
def test_normalize_mdpo_command_aliases(command_aliases, expected_result):
    if hasattr(expected_result, '__traceback__'):
        with pytest.raises(expected_result):
            normalize_mdpo_command_aliases(command_aliases)
    else:
        assert normalize_mdpo_command_aliases(
            command_aliases,
        ) == expected_result


def test_get_valid_mdpo_command_names():
    valid_command_names = get_valid_mdpo_command_names()

    for command_name in valid_command_names:
        assert command_name
        assert isinstance(command_name, str)

        if command_name.startswith('mdpo-'):
            assert removeprefix(command_name, 'mdpo-') in valid_command_names
        else:
            assert f'mdpo-{command_name}' in valid_command_names



"""mdpo I/O utilities."""

import glob
import os


def filter_paths(filepaths, ignore_paths=[]):
    """Filters a list of paths removing those defined in other list of paths.

    The paths to filter can be defined in the list of paths to ignore in
    several forms:

    - The same string.
    - Only the file name.
    - Only their direct directory name.
    - Their direct directory full path.

    Args:
        filepaths (list): Set of source paths to filter.
        ignore_paths (list): Paths that must not be included in the response.

    Returns:
        list: Non filtered paths ordered alphabetically.
    """
    response = []
    for filepath in filepaths:
        # ignore by filename
        if os.path.basename(filepath) in ignore_paths:
            continue
        # ignore by dirname
        if os.path.basename(os.path.dirname(filepath)) in ignore_paths:
            continue
        # ignore by filepath
        if filepath in ignore_paths:
            continue
        # ignore by dirpath (relative or absolute)
        if (os.sep).join(filepath.split(os.sep)[:-1]) in ignore_paths:
            continue
        response.append(filepath)
    response.sort()
    return response


def to_file_content_if_is_file(value, encoding='utf-8'):
    """Check if the value passed is a file path or string content.

    If is a file, reads its content and returns it, otherwise returns
    the string passed as is.

    Args:
        value (str): Value to check if is a filepath or content.
        encoding (str): Expected file encoding, if is a file.

    Returns:
        str: File content if ``value`` is an existing file or ``value`` as is.
    """
    if os.path.isfile(value):
        with open(value, encoding=encoding) as f:
            value = f.read()
    return value


def to_files_or_content(value):
    """File path/glob/content disambiguator.

    Check if the value passed is a glob, a set of files in a list or is string
    content.

    Args:
        value (str): Value to check.

    Returns:
        tuple: Two values being the first a boolean that indicates if ``value``
        is a list of files (``True``) or string content (``False``) and the
        second value is the content, which could be an iterator (if a glob or
        a list of files is passed or a string).
    """
    try:
        parsed = glob.glob(value)
    except TypeError:
        # inferes list
        return (True, value)
    except Exception as err:
        # some strings like '[s-m]' will produce
        # 're.error: bad character range ... at position'

        # NOTE: in Python3.6, 're.error' is 'sre_constants' error,
        # so 'sre_constants' must be removed when Python3.6 support is gone
        if (
            err.__module__ in ['re', 'sre_constants']
            and err.__class__.__name__ == 'error'
        ):
            return (False, value)
        raise err
    if not parsed:
        # assumes it is content
        return (False, value)
    return (True, parsed)


def save_file_checking_file_changed(filepath, content, encoding='utf-8'):
    """Save a file checking if the content has changed.

    Args:
        pofile (:py:class:`polib.POFile`): PO file to save.
        po_filepath (str): Path to the new file to save in.

    Returns:
        bool: If the PO file content has been changed.
    """
    if not os.path.isfile(filepath):
        with open(filepath, 'w', encoding=encoding) as f:
            f.write(content)
        return True

    with open(filepath) as f:
        prev_content = f.read()
    with open(filepath, 'w', encoding=encoding) as f:
        f.write(content)
    return content != prev_content



"""HTML-produced-from-Markdown files translator using PO files as reference."""

import html
import re
import warnings
from collections import OrderedDict
from html.parser import HTMLParser

import md4c
import polib

from mdpo.command import (
    normalize_mdpo_command_aliases,
    parse_mdpo_html_command,
)
from mdpo.io import save_file_checking_file_changed, to_file_content_if_is_file
from mdpo.md import solve_link_reference_targets
from mdpo.po import (
    paths_or_globs_to_unique_pofiles,
    pofiles_to_unique_translations_dicts,
)


PROCESS_REPLACER_TAGS = [
    'p', 'li', 'h1', 'h2', 'h3',
    'h4', 'h5', 'h6', 'td', 'th',
]
ALIGNMENT_CHARS = ['\n', ' ', '\t', '\r']


def _html_attrs_to_str(attrs):
    """Converts a dictionary of HTML attributes to its HTML representation.

    Args:
        attrs (dict): Attributes with values

    Returns:
        str: HTML attributes string ready to be used inside a HTML tag.
    """
    response = ''
    for name, value in attrs.items():
        response += name
        if value is not None:
            response += f'="{value}"'
        response += ' '
    return response.rstrip(' ')


class MdPo2HTML(HTMLParser):
    def __init__(
        self,
        pofiles,
        ignore=[],
        merge_adjacent_markups=False,
        code_tags=['code'],
        bold_tags=['b', 'strong'],
        italic_tags=['em', 'i'],
        link_tags=['a'],
        image_tags=['img'],
        ignore_grouper_tags=['div', 'hr'],
        po_encoding=None,
        command_aliases={},
        _check_saved_files_changed=None,
    ):
        self.pofiles = paths_or_globs_to_unique_pofiles(
            pofiles,
            ignore,
            po_encoding=po_encoding,
        )
        self.output = ''
        self.replacer = []
        self._raw_replacement = ''
        self.context = []
        self._current_msgctxt = None

        self.translations = None
        self.translations_with_msgctxt = None

        self._disable = False
        self._disable_next_line = False
        self._enable_next_line = False
        self.disabled_entries = []

        # custom mdpo command resolution
        self.command_aliases = normalize_mdpo_command_aliases(command_aliases)

        # lazy translators mode
        self.merge_adjacent_markups = merge_adjacent_markups

        # pre-commit mode
        self._saved_files_changed = (
            False if _check_saved_files_changed is not None else None
        )

        # references of msgids and their msgstrs changing referenced link
        # targets by their real target, needed to support link references
        # (this variable is redefined when the first link tag `a` is found)
        self.real_link_reference_targets = None

        # code markup
        self.code_tags = code_tags

        # bold markup
        self.bold_tags = bold_tags

        # italic markup
        self.italic_tags = italic_tags

        # link markup
        self.link_tags = link_tags

        # image markup
        self.image_tags = image_tags

        self.ignore_grouper_tags = ignore_grouper_tags

        self.markup_tags = []
        self.markup_tags.extend(self.code_tags)
        self.markup_tags.extend(self.bold_tags)
        self.markup_tags.extend(self.italic_tags)

        self.html_renderer = md4c.HTMLRenderer(md4c.MD_FLAG_TABLES)

        super().__init__()

    def _merge_adyacent_tags(self, html, template_tags):
        for tags_group in [self.bold_tags, self.italic_tags]:
            regexes = []
            for tag in tags_group:
                if tag not in template_tags:
                    continue
                for _tag in tags_group:
                    if _tag not in template_tags:
                        continue
                    regex = fr'</{tag}>\s*<{_tag}>'
                    if regex not in regexes:
                        regexes.append(regex)

            for regex in regexes:
                html = re.sub(regex, ' ', html)

        return html

    def _remove_lastline_from_output_if_empty(self):
        split_output = self.output.split('\n')
        if not split_output[-1]:
            self.output = '\n'.join(split_output)[:-1]

    def _process_replacer(self):
        # print('REPLACER:', self.replacer)

        template_tags = []
        raw_html_template, _current_replacement = ('', '')

        _current_link_target = ''

        _last_start_tag = None
        _last_end_tag = None
        _inside_code = False
        while self.replacer:
            handle, handled, attrs = self.replacer.pop(0)
            if handle == 'start':
                template_tags.append(handled)

                attrs_repr = (
                    f' {_html_attrs_to_str(attrs)}' if attrs else ''
                )

                if handled in self.code_tags:
                    _current_replacement += '`'
                    raw_html_template += f'<{handled}{attrs_repr}>'
                elif handled in self.bold_tags:
                    _current_replacement += '**'
                    raw_html_template += f'<{handled}{attrs_repr}>'
                elif handled in self.italic_tags:
                    _current_replacement += '*'
                    raw_html_template += f'<{handled}{attrs_repr}>'
                elif handled in self.link_tags:
                    title, href = attrs.get('title'), attrs['href']
                    _current_link_target += f'({href}'
                    if title:
                        _current_link_target += f' "{title}"'
                    _current_link_target += ')'

                    raw_html_template += f'<{handled}'

                    # attrs_except_href_title = []
                    for attr, value in attrs.items():
                        if attr in ['title', 'href']:
                            raw_html_template += f' {attr}="{{}}"'
                        # else:
                        #     These attributes are not included in output
                        #    attrs_except_href_title.append((attr, value))
                    # if attrs_except_href_title:
                    #     raw_html_template += ' '
                    # raw_html_template += _html_attrs_to_str(
                    #     attrs_except_href_title) + '>'
                    raw_html_template += '>'
                else:
                    raw_html_template += f'<{handled}{attrs_repr}>'
                _last_start_tag = handled
                if _last_start_tag == 'code':
                    _inside_code = True

            elif handle == 'data':
                if not _inside_code:
                    while '  ' in handled:
                        handled = handled.replace('  ', ' ')

                if all((ch in ALIGNMENT_CHARS) for ch in handled):
                    raw_html_template += handled
                    _current_replacement += handled
                else:
                    raw_html_template += '{}'
                    if _current_link_target:
                        _current_replacement += (
                            f'[{handled}]{_current_link_target}'
                        )
                        _current_link_target = ''
                    else:
                        _current_replacement += handled
            elif handle == 'end':
                raw_html_template += f'</{handled}>'
                if handled in self.code_tags:
                    _current_replacement += '`'
                elif handled in self.bold_tags:
                    _current_replacement += '**'
                elif handled in self.italic_tags:
                    _current_replacement += '*'
                _last_end_tag = handled
                if _last_end_tag == 'code':
                    _inside_code = False
            elif handle == 'comment':
                raw_html_template += f'<!--{handled}-->'
            elif handle == 'startend':
                if handled in self.image_tags:
                    alt, src = attrs.get('alt', ''), attrs['src']
                    _current_replacement += f'![{alt}]({src}'
                    title = attrs.get('title')
                    if title:
                        _current_replacement += f' "{title}"'
                    _current_replacement += ')'

                    raw_html_template += '{}'
                else:
                    attrs_repr = (
                        f' {_html_attrs_to_str(attrs)}' if attrs else ''
                    )
                    raw_html_template += f'<{handled}{attrs_repr}/>'

        _current_replacement = html.unescape(_current_replacement)

        if (self._disable and not self._enable_next_line) \
                or self._disable_next_line:
            replacement = _current_replacement

            self.disabled_entries.append(
                polib.POEntry(
                    msgid=replacement,
                    msgstr='',
                    msgctxt=self._current_msgctxt,
                ),
            )
        else:
            if self._current_msgctxt:
                replacement = self.translations_with_msgctxt[
                    self._current_msgctxt
                ].get(_current_replacement)
            else:
                replacement = self.translations.get(_current_replacement)
            if not replacement:
                replacement = _current_replacement

        for tt in template_tags:
            for tags_group in [self.bold_tags, self.italic_tags]:
                if tt in tags_group:
                    for tag in tags_group:
                        if tag not in template_tags:
                            template_tags.append(tag)

        # print('RAW TEMPLATE:', raw_html_template)
        # print('TEMPLATE TAGS:', template_tags)
        # print(f'CURRENT MSGID: \'{_current_replacement}\'')
        # print('MSGSTR:', replacement)

        html_before_first_replacement = raw_html_template.split('{')[0]
        html_after_last_replacement = raw_html_template.split('}')[-1]
        for tags_group in [self.bold_tags, self.italic_tags, self.code_tags]:
            for tag in tags_group:
                html_before_first_replacement = \
                    html_before_first_replacement.split(f'<{tag}>')[0]
                html_after_last_replacement = \
                    html_after_last_replacement.split(f'</{tag}>')[-1]
            html_before_first_replacement = \
                html_before_first_replacement.split('<a href="')[0]

        html_inner = '\n'.join(
            self.html_renderer.parse(replacement).split('\n')[:-1],
        )
        html_inner = '>'.join(html_inner.split('>')[1:])
        html_inner = '<'.join(html_inner.split('<')[:-1])

        html_template = html_before_first_replacement + html_inner + \
            html_after_last_replacement

        if self.merge_adjacent_markups:
            html_template = self._merge_adyacent_tags(
                html_template,
                template_tags,
            )

        self.output += html_template
        self.context = []

        self._disable_next_line = False
        self._enable_next_line = False
        self._current_msgctxt = None

        # print('________________________________________________')

    def handle_starttag(self, tag, attrs):
        # print('START TAG: %s | POS: %d:%d' % (tag, *self.getpos()))

        if tag in self.ignore_grouper_tags:
            self.context.append(tag)
            attrs_repr = (
                f' {_html_attrs_to_str(OrderedDict(attrs))}' if attrs else ''
            )
            self.output += f'<{tag}{attrs_repr}>'
        elif (
            (self.context and self.context[0] in self.ignore_grouper_tags) or
            (tag == 'ul' and not self.context) or
            (tag in {'blockquote', 'table', 'thead', 'tbody', 'tr'})
        ):
            attrs_repr = (
                f' {_html_attrs_to_str(OrderedDict(attrs))}' if attrs else ''
            )
            self.output += f'<{tag}{attrs_repr}>'
        else:
            if tag == 'a' and self.real_link_reference_targets is None:
                # extend translations to populate link reference msgid-msgstrs
                # with real targets
                self.real_link_reference_targets = (
                    solve_link_reference_targets(self.translations)
                )
                self.translations.update(self.real_link_reference_targets)

            self.replacer.append(('start', tag, OrderedDict(attrs)))
            self.context.append(tag)

    def handle_endtag(self, tag):
        # print('END TAG: %s | POS: %d:%d' % (tag, *self.getpos()))

        if tag in self.ignore_grouper_tags:
            self.output += f'</{tag}>'
            if self.context:
                self.context.pop()
        elif self.context and self.context[0] in self.ignore_grouper_tags:
            self.output += f'</{tag}>'
        elif tag in PROCESS_REPLACER_TAGS:
            self.replacer.append(('end', tag, None))
            self._process_replacer()
        elif tag in ['ul', 'blockquote', 'tr', 'table', 'thead', 'tbody']:
            self.output += f'</{tag}>'
        else:
            self.replacer.append(('end', tag, None))
            if self.context:
                self.context.pop()

    def handle_startendtag(self, tag, attrs):
        # print('STARTEND TAG: %s | POS: %d:%d' % (tag, *self.getpos()))

        if not self.replacer:
            self.output += self.get_starttag_text()
        else:
            self.replacer.append(('startend', tag, OrderedDict(attrs)))

    def handle_data(self, data):
        # print(f'     DATA: \'{data}\'')

        if data:
            if not self.replacer or (
                    self.context and
                    self.context[0] in self.ignore_grouper_tags
            ) or \
                    not self.context:
                self.output += data
            else:
                if self.context:
                    data = data.replace('\n', ' ')
                self.replacer.append(('data', data, None))

    def handle_comment(self, data):
        # print(f'     COMMENT: \'{data}\'')

        if self.replacer:
            self.replacer.append(('comment', data, None))
        else:
            data_as_comment = f'<!--{data}-->'
            command, comment = parse_mdpo_html_command(
                data_as_comment,
            )
            if command is None:
                self.output += data_as_comment
            else:
                self._remove_lastline_from_output_if_empty()

                try:
                    command = self.command_aliases[command]
                except KeyError:  # not custom command
                    pass

                if command == 'mdpo-disable-next-line':
                    self._disable_next_line = True
                elif command == 'mdpo-disable':
                    self._disable = True
                elif command == 'mdpo-enable':
                    self._disable = False
                elif command == 'mdpo-enable-next-line':
                    self._enable_next_line = True
                elif command == 'mdpo-context' and comment:
                    self._current_msgctxt = comment.strip()
                elif command == 'mdpo-include-codeblock':
                    warnings.warn(
                        'Code blocks translations are not supported'
                        ' by mdpo2html implementation.',
                        SyntaxWarning,
                    )
                else:
                    self.output += data_as_comment

    def translate(self, filepath_or_content, save=None, html_encoding='utf-8'):
        content = to_file_content_if_is_file(
            filepath_or_content,
            encoding=html_encoding,
        )

        self.translations, self.translations_with_msgctxt = (
            pofiles_to_unique_translations_dicts(self.pofiles)
        )

        self.feed(content)

        if save:
            if self._saved_files_changed is False:
                self._saved_files_changed = save_file_checking_file_changed(
                    save,
                    self.output,
                    encoding=html_encoding,
                )
            else:
                with open(save, 'w', encoding=html_encoding) as f:
                    f.write(self.output)

        self.reset()

        return self.output


def markdown_pofile_to_html(
    filepath_or_content,
    pofiles,
    ignore=[],
    save=None,
    po_encoding=None,
    html_encoding='utf-8',
    command_aliases={},
    **kwargs,
):
    r"""Produces a translated HTML file given a previous HTML file (created by a
    Markdown-to-HTML processor) and a set of PO files as reference for msgstrs.

    Args:
        filepath_or_content (str): HTML whose content wants to be translated.
        pofiles (str): Glob for set of pofiles used as reference translating
            the strings to another language.
        ignore (list): List of paths to pofiles to ignore, useful if the glob
            patterns in ``pofiles`` parameter does not fit your requirements.
        save (str): If you pass this parameter as a path to one HTML file,
            even if does not exists, will be saved in the path the output of
            the function.
        html_encoding (str): HTML content encoding.
        po_encoding (str): PO files encoding. If you need different encodings
            for each file, you must define it in the "Content-Type" field of
            each PO file metadata, in the form
            ``"Content-Type: text/plain; charset=<ENCODING>\n"``.
        command_aliases (dict): Mapping of aliases to use custom mdpo command
            names in comments. The ``mdpo-`` prefix in command names resolution
            is optional. For example, if you want to use ``<!-- mdpo-on -->``
            instead of ``<!-- mdpo-enable -->``, you can pass the dictionaries
            ``{"mdpo-on": "mdpo-enable"}`` or ``{"mdpo-on": "enable"}`` to this
            parameter.

    .. rubric:: Known limitations:

    * This implementation doesn't include support for
      :ref:`include-codeblock command<include-codeblock-command>`.


    Returns:
        str: HTML output translated version of the given file.
    """
    return MdPo2HTML(
        pofiles,
        ignore=ignore,
        po_encoding=po_encoding,
        command_aliases=command_aliases,
        **kwargs,
    ).translate(
        filepath_or_content,
        save=save,
        html_encoding=html_encoding,
    )



"""md4c related stuff for mdpo."""


#: :list: `md4c parser <https://github.com/mity/md4c>`_ extensions used as
#: default by :ref:`md2po<md2po-init>` and :ref:`po2md<po2md-init>`
#: implementations.
DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS = [
    'collapse_whitespace',
    'tables',
    'strikethrough',
    'tasklists',
    'latex_math_spans',
    'wikilinks',
]

#: :dict: mapping from `pymd4c <https://github.com/dominickpastore/pymd4c>`_
#: block type values to block readable names
READABLE_BLOCK_NAMES = {
    1: 'quote',
    2: 'unordered list',
    3: 'ordered list',
    6: 'header',
    7: 'code',
    8: 'html',
    9: 'paragraph',
    10: 'table',
}



"""Markdown to PO file to Markdown translator."""

import glob
import os

from mdpo.md2po import Md2Po
from mdpo.md4c import DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS
from mdpo.po2md import Po2Md


def markdown_to_pofile_to_markdown(
    langs,
    input_paths_glob,
    output_paths_schema,
    extensions=DEFAULT_MD4C_GENERIC_PARSER_EXTENSIONS,
    command_aliases={},
    location=True,
    debug=False,
    po_encoding=None,
    md_encoding=None,
    md2po_kwargs={},
    po2md_kwargs={},
    _check_saved_files_changed=False,
):
    """Translate a set of Markdown files using PO files.

    Args:
        langs (list): List of languages used to build the output directories.
        input_paths_glob (str): Glob covering Markdown files to translate.
        output_paths_schema (str): Path schema for outputs, built using
            placeholders. There is a mandatory placeholder for languages:
            ``{lang}``; and one optional for output basename: ``{basename}``.
            For example, for the schema ``locale/{lang}``, the languages
            ``['es', 'fr']`` and a ``README.md`` as input, the next files will
            be written:

            * ``locale/es/README.po``
            * ``locale/es/README.md``
            * ``locale/fr/README.po``
            * ``locale/fr/README.md``

            Note that you can omit ``{basename}``, specifying a directory for
            each language with ``locale/{lang}`` for this example.
            Unexistent directories and files will be created, so you don't
            have to prepare the output directories before the execution.
        extensions (list): md4c extensions used to parse markdown content,
            formatted as a list of 'pymd4c' keyword arguments. You can see all
            available at `pymd4c repository <https://github.com/dominickpastore
            /pymd4c#parser-option-flags>`_.
        command_aliases (dict): Mapping of aliases to use custom mdpo command
            names in comments. The ``mdpo-`` prefix in command names resolution
            is optional. For example, if you want to use ``<!-- mdpo-on -->``
            instead of ``<!-- mdpo-enable -->``, you can pass the dictionaries
            ``{"mdpo-on": "mdpo-enable"}`` or ``{"mdpo-on": "enable"}`` to this
            parameter.
        location (bool): Store references of top-level blocks in which are
            found the messages in PO file `#: reference` comments.
        debug (bool): Add events displaying all parsed elements in the
            extraction process.
        po_encoding (str): PO files encoding.
        md_encoding (str): Markdown files encoding.
        md2po_kwargs (dict): Additional optional arguments passed to
            ``markdown_to_pofile`` function.
        po2md_kwargs (dict): Additional optional arguments passed to
            ``pofile_to_markdown`` function.
    """
    if '{lang}' not in output_paths_schema:
        raise ValueError(
            "You must pass the replacer '{lang}' inside the argument"
            " 'output_paths_schema'.",
        )

    try:
        input_paths_glob_ = glob.glob(input_paths_glob)
    except Exception as err:
        if (
            err.__module__ in ['re', 'sre_constants']
            and err.__class__.__name__ == 'error'
        ):
            # some strings like '[s-m]' will produce
            # 're.error: bad character range ... at position'
            raise ValueError(
                "The argument 'input_paths_glob' must be a valid glob or file"
                ' path.',
            )
        raise err
    else:
        if not input_paths_glob_:
            raise FileNotFoundError(
                f'The glob \'{input_paths_glob}\' does not match any file.',
            )

    _saved_files_changed = None if not _check_saved_files_changed else False

    for filepath in input_paths_glob_:
        for lang in langs:
            md_ext = os.path.splitext(filepath)[-1]

            file_basename = os.path.splitext(os.path.basename(filepath))[0]

            format_kwargs = {'lang': lang}
            if '{basename}' in output_paths_schema:
                format_kwargs['basename'] = file_basename
            po_filepath = output_paths_schema.format(**format_kwargs)

            po_basename = os.path.basename(po_filepath)
            po_dirpath = (
                os.path.dirname(po_filepath)
                if (po_basename.count('.') or file_basename == po_basename)
                else po_filepath
            )

            os.makedirs(os.path.abspath(po_dirpath), exist_ok=True)
            if os.path.isdir(po_filepath):
                po_filepath = os.path.join(
                    po_filepath.rstrip(os.sep),
                    f'{os.path.basename(filepath)}.po',
                )
            if not po_filepath.endswith('.po'):
                po_filepath += '.po'

            format_kwargs['ext'] = md_ext.lstrip('.')
            md_filepath = output_paths_schema.format(**format_kwargs)
            if os.path.isdir(md_filepath):
                md_filepath = (
                    md_filepath.rstrip(os.sep) + os.sep +
                    os.path.basename(filepath)
                )

            # md2po
            md2po = Md2Po(
                filepath,
                extensions=extensions,
                command_aliases=command_aliases,
                debug=debug,
                location=location,
                _check_saved_files_changed=_check_saved_files_changed,
                **md2po_kwargs,
            )
            md2po.extract(
                save=True,
                po_filepath=po_filepath,
                po_encoding=po_encoding,
                md_encoding=md_encoding,
            )
            if _check_saved_files_changed and _saved_files_changed is False:
                _saved_files_changed = md2po._saved_files_changed

            # po2md
            po2md = Po2Md(
                [po_filepath],
                command_aliases=command_aliases,
                debug=debug,
                po_encoding=po_encoding,
                _check_saved_files_changed=_check_saved_files_changed,
                **po2md_kwargs,
            )
            po2md.translate(
                filepath,
                save=md_filepath,
                md_encoding=md_encoding,
            )
            if _check_saved_files_changed and _saved_files_changed is False:
                _saved_files_changed = po2md._saved_files_changed

    return _saved_files_changed
